"""
Routines for extracting data from Siemens DICOM files.

The simplest way to read a file is to call read(filename). If you like you
can also call lower level functions like read_data().

Except for the map of internal data types to numpy type strings (which
doesn't require an import of numpy), this code is deliberately ignorant of
numpy. It returns native Python types that are easy to convert into
numpy types.
"""

# Python modules
from __future__ import division
import struct
import exceptions
import math

# 3rd party modules
import dicom

# Our modules
import util_mrs_file
import constants

TYPE_NONE = 0
TYPE_IMAGE = 1
TYPE_SPECTROSCOPY = 2


# Change to True to enable the assert() statements sprinkled through the code
ASSERTIONS_ENABLED = False

# THese are some Siemens-specific tags
TAG_CONTENT_TYPE = (0x0029, 0x1008)
TAG_SPECTROSCOPY_DATA = (0x7fe1, 0x1010)

# I (Philip) ported much of the private tag parsing code from the IDL routines
# dicom_fill_rsp.pro and dicom_fill_util.pro, except for the CSA header
# parsing which is a port of C++ code in the GDCM project.

# Since a lot (all?) of the Siemens format is undocumented, there are magic
# numbers and logic in here that I can't explain. Sorry! Where appropriate
# I have copied or paraphrased comments from the IDL code; they're marked
# with [IDL]. Unmarked comments are mine. Where ambiguous, I labelled my
# comments with [PS] (Philip Semanchuk).


def read(filename, ignore_data=False):
    """ This is the simplest (and recommended) way for our code to read a
    Siemens DICOM file.
    
    It returns a tuple of (parameters, data). The parameters are a dict.
    The data is in a Python list.
    """
    # Since a DICOM file is params + data together, it's not so simple to
    # ignore the data part. The best we can do is tell PyDicom to apply
    # lazy evaluation which is probably less efficient in the long run.
    defer_size = 4096 if ignore_data else 0
    
    dataset = dicom.read_file(filename)
    
    params = read_parameters_from_dataset(dataset)
    data = read_data_from_dataset(dataset)
    
    return params, data
    
    
def read_parameters(filename):
    return read_parameters_from_dataset(dicom.read_file(filename))


def read_data(filename):
    return read_data_from_dataset(dicom.read_file(filename))

    
def read_data_from_dataset(dataset):
    """Given a PyDicom dataset, returns the data in the Siemens DICOM 
    spectroscopy data tag (0x7fe1, 0x1010) as a list of complex numbers.    
    """
    data = _get(dataset, TAG_SPECTROSCOPY_DATA)
    if data:
        # Big simplifying assumptions --
        # 1) Data is a series of complex numbers organized as ririri... 
        #    where r = real and i = imaginary.
        # 2) Each real & imaginary number is a 4 byte float.
        # 3) Data is little endian.
        data = struct.unpack("<%df" % (len(data) / 4), data)
        data = util_mrs_file.collapse_complexes(data)
    else:
        data = [ ]
        
    return data


def read_parameters_from_dataset(dataset):
    """Given a PyDicom dataset, returns a fairly extensive subset of the
    parameters therein as a dictionary.
    """
    params = { }
    
    # The code below refers to slice_index as a variable, but here it is 
    # hardcoded to one. It could vary, in theory, but in practice I don't
    # know how it would actually be used. How would the slice index or 
    # indices be passed? How would the data be returned? For now, I'll 
    # leave the slice code active but hardcode the index to 1.
    slice_index = 1
    
    # [PS] - Even after porting this code I still can't figure out what
    # ptag_img and ptag_ser stand for, so I left the names as is.
    ptag_img = { }
    ptag_ser = { }

    # (0x0029, 0x__10) is one of several possibilities
    # - SIEMENS CSA NON-IMAGE, CSA Data Info
    # - SIEMENS CSA HEADER, CSA Image Header Info
    # - SIEMENS CSA ENVELOPE, syngo Report Data
    # - SIEMENS MEDCOM HEADER, MedCom Header Info
    # - SIEMENS MEDCOM OOG, MedCom OOG Info (MEDCOM Object Oriented Graphics)
    # Pydicom identifies it as "CSA Image Header Info"
    for tag in ( (0x0029, 0x1010), (0x0029, 0x1210), (0x0029, 0x1110) ):
        tag_data = dataset.get(tag, None)
        if tag_data:
            break

    if tag_data:
        ptag_img = _parse_csa_header(tag_data.value)

    # [IDL] Access the SERIES Shadow Data
    # [PS] I don't know what makes this "shadow" data.
    for tag in ( (0x0029, 0x1020), (0x0029, 0x1220), (0x0029, 0x1120) ):
        tag_data = dataset.get(tag, None)
        if tag_data:
            break

    if tag_data:
        ptag_ser = _parse_csa_header(tag_data.value)

    # [IDL] "MrProtocol" (VA25) and "MrPhoenixProtocol" (VB13) are special
    # elements that contain many parameters.
    if ptag_ser.get("MrProtocol", ""):
        prot_ser = _parse_protocol_data(ptag_ser["MrProtocol"])

    if ptag_ser.get("MrPhoenixProtocol", ""):
        prot_ser = _parse_protocol_data(ptag_ser["MrPhoenixProtocol"])

    # [IDL] Determine if file is SVS,SI,EPSI, or OTHER
    # [PS] IDL code doesn't match comments. Possibilities appear to
    #      include EPSI, SVS, CSI, JPRESS and SVSLIP2. "OTHER" isn't
    #      considered.
    #      EPSI = Echo-Planar Spectroscopic Imaging
    #      SVS = Single voxel spectroscopy
    #      CSI = Chemical Shift Imaging
    #      JPRESS = J-resolved spectroscopy
    #      SVSLIP2 = No idea!
    is_epsi = False
    is_svs = False
    is_csi = False
    is_jpress = False
    is_svslip2 = False

    # [IDL] Protocol name
    parameter_filename = _extract_from_quotes(prot_ser.get("tProtocolName", ""))
    parameter_filename = parameter_filename.strip()

    # [IDL] Sequence file name
    sequence_filename = _extract_from_quotes(prot_ser.get("tSequenceFileName", ""))
    sequence_filename = sequence_filename.strip()

    sequence_filename2 = ptag_img.get("SequenceName", "")
    sequence_filename2 = sequence_filename2.strip()

    parameter_filename_lower = parameter_filename.lower()
    sequence_filename_lower = sequence_filename.lower()
    sequence_filename2_lower = sequence_filename2.lower()

    is_epsi = ("epsi" in (parameter_filename_lower, sequence_filename_lower))

    is_svs = ("svs" in (parameter_filename_lower, sequence_filename_lower,
                        sequence_filename2_lower))

    if "fid" in (parameter_filename_lower, sequence_filename_lower):
        if "csi" in (parameter_filename_lower, sequence_filename_lower):
            is_csi = True
        else:
            is_svs = True

    if "csi" in (parameter_filename_lower, sequence_filename_lower):
        is_csi = True

    is_jpress = ("jpress" in (parameter_filename_lower,
                              sequence_filename_lower))

    is_svslip2 = ("svs_li2" in (parameter_filename_lower,
                                sequence_filename2_lower))


    # Patient Info
    params["patient_name"] = _get(dataset, (0x0010, 0x0010), "")
    params["patient_id"] = _get(dataset, (0x0010, 0x0020))
    params["patient_birthdate"] = _get(dataset, (0x0010, 0x0030))
    params["patient_sex"] = _get(dataset, (0x0010, 0x0040), "")
    # [PS] Siemens stores the age as nnnY where 'n' is a digit, e.g. 042Y
    params["patient_age"] = \
                        int(_get(dataset, (0x0010, 0x1010), "000Y")[:3])
    params["patient_weight"] = round(_get(dataset, (0x0010, 0x1030), 0))
    params["study_code"] = _get(dataset, (0x0008, 0x1030), "")

    # Identification info
    params["bed_move_fraction"] = 0.0

    s = _get(dataset, (0x0008, 0x0080), "")
    if s:
        s = " " + s
    s += _get(dataset, (0x0008, 0x1090), "")
    params["institution_id"] = s
    params["parameter_filename"] = parameter_filename
    params["study_type"] = "spec"

    # DICOM date format is YYYYMMDD
    params["bed_move_date"] = _get(dataset, (0x0008, 0x0020), "")
    params["measure_date"] = params["bed_move_date"]
    # DICOM time format is hhmmss.fraction
    params["bed_move_time"] = _get(dataset, (0x0008, 0x0030), "")
    params["comment_1"] = _get(dataset, (0x0008, 0x0031), "")
    if not params["comment_1"]:
        params["comment_1"] = _get(dataset, (0x0020, 0x4000), "")
    # DICOM time format is hhmmss.fraction
    params["measure_time"] = _get(dataset, (0x0008, 0x0032), "")

    params["sequence_filename"] = ptag_img.get("SequenceName", "")
    params["sequence_type"] = ptag_img.get("SequenceName", "")

    # Measurement info
    params["echo_position"] = "0.0"
    params["image_contrast_mode"] = "unknown"
    params["kspace_mode"] = "unknown"
    params["measured_slices"] = "1"
    params["saturation_bands"] = "0"

    # Seems to me that a quantity called "NumberOfAverages" would be an
    # int, but it is stored as a float, e.g. "128.0000" which makes
    # Python's int() choke unless I run it through float() first.
    params["averages"] = int(_float(ptag_img.get("NumberOfAverages", "")))
    params["flip_angle"] = _float(ptag_img.get("FlipAngle", ""))
    # [PS] DICOM stores frequency as MHz, we store it as Hz. Mega = 1x10(6)
    params["frequency"] = float(ptag_img.get("ImagingFrequency", 0)) * 1e6

    inversion_time = float(ptag_img.get("InversionTime", 0))
    params["inversion_time_1"] = inversion_time
    params["number_inversions"] = 1 if inversion_time else 0

    params["measured_echoes"] = ptag_img.get("EchoTrainLength", "1")
    params["nucleus"] = ptag_img.get("ImagedNucleus", "")
    params["prescans"] = prot_ser.get("sSpecPara.lPreparingScans", 0)

    # Gain
    gain = prot_ser.get("sRXSPEC.lGain", None)
    if gain == 0:
        gain = "-20.0"
    elif gain == 1:
        gain = "0.0"
    else:
        gain = ""

    params["receiver_gain"] = gain

    params["ft_scale_factor"] = \
                float(prot_ser.get("sRXSPEC.aFFT_SCALE[0].flFactor", 0))

    # Receiver Coil
    coil = prot_ser.get("sCOIL_SELECT_MEAS.asList[0].sCoilElementID.tCoilID", "")
    params["receiver_coil"] = _extract_from_quotes(coil)

    # [IDL] differs in EPSI
    params["repetition_time_1"] = float(prot_ser.get("alTR[0]", 0)) * 0.001

    sweep_width = ""
    remove_oversample_flag = prot_ser.get("sSpecPara.ucRemoveOversampling", "")
    remove_oversample_flag = (remove_oversample_flag.strip() == "0x1")
    readout_os = float(ptag_ser.get("ReadoutOS", 1.0))

    dwelltime = float(ptag_img.get("RealDwellTime", 1.0)) * 1e-9

    if dwelltime:
        sweep_width = 1 / dwelltime
        if not remove_oversample_flag:
            sweep_width *= readout_os

        sweep_width = str(sweep_width)

    params["transmitter_voltage"] = \
        prot_ser.get("sTXSPEC.asNucleusInfo[0].flReferenceAmplitude", "0.0")
    params["total_duration"] = \
        prot_ser.get("lTotalScanTimeSec", "0.0")

    prefix = "sSliceArray.asSlice[%d]." % slice_index
    image_parameters = (
                        ("image_dimension_line",      "dPhaseFOV"),
                        ("image_dimension_column",    "dReadoutFOV"),
                        ("image_dimension_partition", "dThickness"),
                        ("image_position_sagittal",   "sPosition.dSag"),
                        ("image_position_coronal",    "sPosition.dCor"),
                        ("image_position_transverse", "sPosition.dTra"),
                       )
    for key, name in image_parameters:
        params[key] = float(prot_ser.get(prefix + name, "0.0"))

    # [IDL] Image Normal/Column
    image_orientation = ptag_img.get("ImageOrientationPatient", "")

    if not image_orientation:
        slice_orientation_pitch = ""
        slice_distance = ""
    else:
        # image_orientation is a list of strings, e.g. --
        # ['-1.00000000', '0.00000000', '0.00000000', '0.00000000', 
        #   '1.00000000', '0.00000000']

        # [IDL] If the data we are processing is a Single Voxel
        # Spectroscopy data, interchange rows and columns. Due to an error
        # in the protocol used.
        if is_svs:
            image_orientation = image_orientation[3:] + image_orientation[:3]

        # Convert the values to float and discard ones smaller than 1e-4
        f = lambda value: 0.0 if abs(value) < 1e-4 else value
        image_orientation = [f(float(value)) for value in image_orientation]

        row = image_orientation[:3]
        column = image_orientation[3:6]
        normal = ( ((row[1] * column[2]) - (row[2] * column[1])), 
                   ((row[2] * column[0]) - (row[0] * column[2])),
                   ((row[0] * column[1]) - (row[1] * column[0])),
                 )

        params["image_normal_sagittal"] = normal[0]
        params["image_normal_coronal"] = normal[1]
        params["image_normal_transverse"] = normal[2]
        params["image_column_sagittal"] = column[0]
        params["image_column_coronal"] = column[0]
        params["image_column_transverse"] = column[0]

        # Second part of the return tuple is orientation; we don't use it.
        slice_orientation_pitch, _ = _dicom_orientation_string(normal)

        # Slice distance
        # http://en.wikipedia.org/wiki/Dot_product
        keys = ("image_position_sagittal", "image_position_coronal",
                "image_position_transverse")
        a = [params[key] for key in keys]
        b = normal
        bb = math.sqrt(sum([value ** 2 for value in normal]))

        slice_distance = ((a[0] * b[0]) + (a[1] * b[1]) + (a[2] * b[2])) / bb


    params["slice_orientation_pitch"] = slice_orientation_pitch
    params["slice_distance"] = slice_distance

    regions = ( ("region_dimension_line",      "dPhaseFOV"),
                ("region_dimension_column",    "dReadoutFOV"),
                ("region_dimension_partition", "dThickness"),
                ("region_position_sagittal",   "sPosition.dSag"),
                ("region_position_coronal",    "sPosition.dCor"),
                ("region_position_transverse", "sPosition.dTra"),
              )

    for key, name in regions:
        name = "sSpecPara.sVoI." + name
        params[key] = float(prot_ser.get(name, 0))

    # 'DATA INFORMATION'

    params["measure_size_spectral"] = \
                                long(prot_ser.get('sSpecPara.lVectorSize', 0))

    params["slice_thickness"] = _float(ptag_img.get("SliceThickness", 0))
    params["current_slice"] = "1"
    params["number_echoes"] = "1"
    params["number_slices"] = "1"
    params["data_size_spectral"] = params["measure_size_spectral"]

    # ;------------------------------------------------------
    # [IDL] Sequence Specific Changes
    if not is_epsi:
        # [IDL] Echo time - JPRESS handling added by Dragan
        echo_time = 0.0

        if is_jpress:
            # [IDL] Yingjian saves echo time in a private 'echotime' field
            # [PS] The IDL code didn't use a dict to store these values
            # but instead did a brute force case-insensitive search over
            # an array of strings. In that context, key case didn't matter
            # but here it does.
            keys = prot_ser.keys()
            for key in keys:
                if key.upper() == "ECHOTIME":
                    echo_time = float(prot_ser[key])

        if is_svslip2:
            # [IDL] BJS found TE value set in ICE to be updated in
            #      'echotime' field
            # [PS] The IDL code didn't use a dict to store these values
            # but instead did a brute force case-insensitive search over
            # an array of strings. In that context, key case didn't matter
            # but here it does.
            keys = ptag_img.keys()
            for key in keys:
                if key.upper() == "ECHOTIME":
                    echo_time = float(ptag_img[key])

        if not echo_time:
            # [IDL] still no echo time - try std place
            echo_time = float(prot_ser.get('alTE[0]', 0.0))
            echo_time /= 1000

        params["echo_time"] = echo_time

        params["data_size_line"] = \
                    int(prot_ser.get('sSpecPara.lFinalMatrixSizePhase', 1))

        params["data_size_column"] = \
                        int(prot_ser.get('sSpecPara.lFinalMatrixSizeRead', 1))

        params["data_size_partition"] = \
                    int(prot_ser.get('sSpecPara.lFinalMatrixSizeSlice', 1))

        if is_svs:
            # [IDL] For Single Voxel Spectroscopy data (SVS) only
            params["image_dimension_line"] = \
                                        params["region_dimension_line"]
            params["image_dimension_column"] = \
                                        params["region_dimension_column"]
            params["image_dimension_partition"] = \
                                        params["region_dimension_partition"]


            # [IDL] For SVS data the following three parameters cannot be
            # anything other than 1
            params["measure_size_line"] = 1
            params["measure_size_column"] = 1
            params["measure_size_partition"] = 1
        else:
            # Not SVS
            #   ;--------------------------------------------------
            #   ; [IDL] For CSI or OTHER Spectroscopy data only
            #   ;--------------------------------------------------

            measure_size_line = int(prot_ser.get('sKSpace.lPhaseEncodingLines', 1))
            params["measure_size_line"] = str(measure_size_line)

            measure_size_column = int(prot_ser.get('sKSpace.lPhaseEncodingLines', 0))
            params["measure_size_column"] = str(measure_size_column)

            measure_size_partition = int(prot_ser.get('sKSpace.lPartitions', '0'))
            kspace_dimension = prot_ser.get('sKSpace.ucDimension', '')
            if kspace_dimension.strip() == "0x2":
                measure_size_partition = 1
                params["data_size_partition"] = 1
                data_size_partition = 1

            params["measure_size_partition"] = measure_size_partition


    if sequence_filename in ("svs_cp_press", "svs_se_ir", "svs_tavg"):
        # [IDL] Inversion Type 0-Volume,1-None
        s = prot_ser.get("SPREPPULSES.UCINVERSION", "")
        if s == "0x1":
            params["number_inversions"] = 1
        elif s == "0x2":
            params["number_inversions"] = 0
        # else:
        #     params["number_inversions"] doesn't get set at all.
        #     This matches the behavior of the IDL code. Note that
        #     params["number_inversions"] is also populated
        #     unconditionally in code many lines above.


    if sequence_filename in ("svs_se", "svs_st", "fid", "fid3", "fid_var",
                             "csi_se", "csi_st", "csi_fid", "csi_fidvar",
                             "epsi"):
         # [IDL] FOR EPSI Measure_size and Data_size parameters are the same
        params["region_dimension_line"] = \
                                params["image_dimension_line"]
        params["region_dimension_column"] = \
                                params["image_dimension_column"]
        params["ft_scale_factor"] = "1.0"

        params["data_size_line"] = \
                int(prot_ser.get('sKSpace.lPhaseEncodingLines', 0))
        params["data_size_column"] = \
                int(prot_ser.get('sKSpace.lBaseResolution', 0)) * readout_os
        params["data_size_partition"] = \
                int(prot_ser.get('sKSpace.lPartitions', 0))

        params["measure_size_line"] = params["data_size_line"]
        measure_size_column = params["data_size_column"]
        measure_size_partition = params["data_size_partition"]

        index = 0 if ((int(dataset.get("InstanceNumber", 0)) % 2) == 1) else 1
        echo_time =         float(prot_ser.get('alTE[%d]' % index, 0)) / 1000
        repetition_time_1 = float(prot_ser.get('alTR[%d]' % index, 0)) / 1000

        params["echo_time"] = str(echo_time)
        params["repetition_time_1"] = str(repetition_time_1)

        dwelltime = float(ptag_img.get("RealDwellTime", 0.0))
        if dwelltime and base_resolution:
            sweep_width = 1 / (dwelltime * base_resolution * readout_os)
        else:
            sweep_width = ""

        params["sweep_width"] = sweep_width

    # Added by BTA
    ip_rot = prot_ser.get("sSliceArray.asSlice[0].dInPlaneRot", None)
    pol_swap = prot_ser.get("sWipMemBlock.alFree[40]", None)
    if ip_rot:
        try:
            ip_rot = float(ip_rot)
            params["in_plane_rotation"] = ip_rot
        except Exception, e:
            pass
    if pol_swap:
        try:
            pol_swap = int(pol_swap)
            params["polarity_swap"] = pol_swap
        except Exception, e:
            raise e

    return params


def _my_assert(expression):
    if ASSERTIONS_ENABLED:
        assert(expression)


def _dicom_orientation_string(normal):
    """Given a 3-item list (or other iterable) that represents a normal vector
    to the "imaging" plane, this function determines the orientation of the
    vector in 3-dimensional space. It returns a tuple of (angle, orientation)
    in which angle is e.g. "Tra" or "Tra>Cor -6" or "Tra>Sag 14.1 >Cor 9.3"
    and orientation is e.g. "Sag" or "Cor-Tra".

    For double angulation, errors in secondary angle occur that may be due to
    rounding errors in internal Siemens software, which calculates row and
    column vectors.
    """
    # docstring paraphrases IDL comments
    TOLERANCE = 1.e-4
    orientations = ('Sag', 'Cor', 'Tra')

    final_angle = ""
    final_orientation = ""

    # [IDL] evaluate orientation of normal vector:
    #
    # Find principal direction of normal vector (i.e. axis with its largest
    # component)
    # Find secondary direction (second largest component)
    # Calc. angle btw. projection of normal vector into the plane that
    #     includes both principal and secondary directions on the one hand
    #     and the principal direction on the other hand ==> 1st angulation:
    #     "principal>secondary = angle"
    # Calc. angle btw. projection into plane perpendicular to principal
    #     direction on the one hand and secondary direction on the other
    #     hand ==> 2nd angulation: "secondary>third dir. = angle"


    # get principal, secondary and ternary directions
    sorted_normal = sorted(normal)

    for i, value in enumerate(normal):
        if value == sorted_normal[2]:
            # [IDL] index of principal direction
            principal = i
        if value == sorted_normal[1]:
            # [IDL] index of secondary direction
            secondary = i
        if value == sorted_normal[0]:
            # [IDL] index of ternary direction
            ternary = i

    # [IDL] calc. angle between projection into third plane (spawned by
    # principle & secondary directions) and principal direction:
    angle_1 = math.atan2(normal[secondary], normal[principal]) * \
                                                constants.RADIANS_TO_DEGREES

    # [IDL] calc. angle btw. projection on rotated principle direction and
    # secondary direction:
    # projection on rotated principle dir.
    new_normal_ip = math.sqrt((normal[principal] ** 2) + (normal[secondary] ** 2))

    angle_2 = math.atan2(normal[ternary], new_normal_ip) * \
                                                constants.RADIANS_TO_DEGREES

    # [IDL] SIEMENS notation requires modifications IF principal dir. indxs SAG !
    # [PS] In IDL, indxs is the name of the variable that is "secondary" here.
    #      Even with that substitution, I don't understand the comment above.
    if not principal:
        if abs(angle_1) > 0:
            sign1 = angle_1 / abs(angle_1)
        else:
            sign1 = 1.0

        angle_1 -= (sign1 * 180.0)
        angle_2 *= -1

    if (abs(angle_2) < TOLERANCE) or (abs(abs(angle_2) - 180) < TOLERANCE):
        if (abs(angle_1) < TOLERANCE) or (abs(abs(angle_1) - 180) < TOLERANCE):
            # [IDL] NON-OBLIQUE:
            final_angle = orientations[principal]
            final_orientation = ang
        else:
            # [IDL] SINGLE-OBLIQUE:
            final_angle = "%s>%s %.3f" % \
                    (orientations[principal], orientations[secondary],
                     (-1 * angle_1)
                    )
            final_orientation = orientations[principal] + '-' + orientations[secondary]
    else:
        # [IDL] DOUBLE-OBLIQUE:
        final_angle = "%s>%s %.3f >%s %f" % \
                (orientations[principal], orientations[secondary],
                 (-1 * angle_1), orientations[ternary], (-1 * angle_2))
        final_orientation = "%s-%s-%s" % \
                (orientations[principal], orientations[secondary],
                 orientations[ternary])

    return final_angle, final_orientation


def _float(value):
    """Attempts to return value as a float. No different from Python's 
    built-in float(), except that it accepts None and "" (for which it 
    returns 0.0).
    """
    return float(value) if value else 0.0
    
    
def _extract_from_quotes(s):
    """Given a string, returns the portion between the first and last
    double quote (ASCII 34). If there aren't at least two quote characters,
    the original string is returned."""
    start = s.find('"')
    end = s.rfind('"')

    if (start != -1) and (end != -1):
        s = s[start + 1 : end]

    return s


def _null_truncate(s):
    """Given a string, returns a version truncated at the first '\0' if
    there is one. If not, the original string is returned."""
    i = s.find(chr(0))
    if i != -1:
        s = s[:i]

    return s


def _scrub(item):
    """Given a string, returns a version truncated at the first '\0' and
    stripped of leading/trailing whitespace. If the param is not a string,
    it is returned unchanged."""
    if isinstance(item, basestring):
        return _null_truncate(item).strip()
    else:
        return item


def _get_chunks(tag, index, format, little_endian=True):
    """Given a CSA tag string, an index into that string, and a format
    specifier compatible with Python's struct module, returns a tuple
    of (size, chunks) where size is the number of bytes read and
    chunks are the data items returned by struct.unpack(). Strings in the
    list of chunks have been run through _scrub().
    """
    # The first character of the format string indicates endianness.
    format = ('<' if little_endian else '>') + format
    size = struct.calcsize(format)
    chunks = struct.unpack(format, tag[index:index + size])

    chunks = [_scrub(item) for item in chunks]

    return (size, chunks)


def _parse_protocol_data(protocol_data):
    """Returns a dictionary containing the name/value pairs inside the
    "ASCCONV" section of the MrProtocol or MrPhoenixProtocol elements
    of a Siemens CSA Header tag.
    """
    # Protocol_data is a large string (e.g. 32k) that lists a lot of
    # variables in a JSONish format with which I'm not familiar. Following
    # that there's another chunk of data delimited by the strings you see
    # below.
    # That chunk is a list of name=value pairs, INI file style. We
    # ignore everything outside of the ASCCONV delimiters. Everything inside
    # we parse and return as a dictionary.
    start = protocol_data.find("### ASCCONV BEGIN ###")
    end = protocol_data.find("### ASCCONV END ###")

    _my_assert(start != -1)
    _my_assert(end != -1)

    start += len("### ASCCONV BEGIN ###")
    protocol_data = protocol_data[start:end]

    lines = protocol_data.split('\n')

    # The two lines of code below turn the 'lines' list into a list of
    # (name, value) tuples in which name & value have been stripped and
    # all blank lines have been discarded.
    f = lambda pair: (pair[0].strip(), pair[1].strip())
    lines = [f(line.split('=')) for line in lines if '=' in line]

    return dict(lines)


def _get(dataset, tag, default=None):
    """Returns the value of a dataset tag, or the default if the tag isn't
    in the dataset.
    PyDicom datasets already have a .get() method, but it returns a
    dicom.DataElement object. In practice it's awkward to call dataset.get()
    and then figure out if the result is the default or a DataElement,
    and if it is the latter _get the .value attribute. This function allows
    me to avoid all that mess.
    It is also a workaround for this bug (which I submitted) which should be
    fixed in PyDicom > 0.9.3:
    http://code.google.com/p/pydicom/issues/detail?id=72
    """
    return default if tag not in dataset else dataset[tag].value


def _parse_csa_header(tag, little_endian = True):
    """The CSA header is a Siemens private tag that should be passed as
    a string. Any of the following tags should work: (0x0029, 0x1010),
    (0x0029, 0x1210), (0x0029, 0x1110), (0x0029, 0x1020), (0x0029, 0x1220),
    (0x0029, 0x1120).

    The function returns a dictionary keyed by element name.
    """
    # Let's have a bit of fun, shall we? A Siemens CSA header is a mix of
    # binary glop, ASCII, binary masquerading as ASCII, and noise masquerading
    # as signal. It's also undocumented, so there's no specification to which
    # to refer.

    # The format is a good one to show to anyone who complains about XML being
    # verbose or hard to read. Spend an afternoon with this and XML will
    # look terse and read like a Shakespearean sonnet.

    # The algorithm below is a translation of the GDCM project's
    # CSAHeader::LoadFromDataElement() inside gdcmCSAHeader.cxx. I don't know
    # how that code's author figured out what's in a CSA header, but the
    # code works.

    # I added comments and observations, but they're inferences. I might
    # be wrong. YMMV.

    # Some observations --
    # - If you need to debug this code, a hexdump of the tag data will be
    #   your best friend.
    # - The data in the tag is a list of elements, each of which contains
    #   zero or more subelements. The subelements can't be further divided
    #   and are either empty or contain a string.
    # - Everything begins on four byte boundaries.
    # - This code will break on big endian data. I don't know if this data
    #   can be big endian, and if that's possible I don't know what flag to
    #   read to indicate that. However, it's easy to pass an endianness flag
    #   to _get_chunks() should the need to parse big endian data arise.
    # - Delimiters are thrown in here and there; they are 0x4d = 77 which is
    #   ASCII 'M' and 0xcd = 205 which has no ASCII representation.
    # - Strings in the data are C-style NULL terminated.

    # I sometimes read delimiters as strings and sometimes as longs.
    DELIMITERS = ("M", "\xcd", 0x4d, 0xcd)

    # This dictionary of elements is what this function returns
    elements = { }

    # I march through the tag data byte by byte (actually a minimum of four
    # bytes at a time), and current points to my current position in the tag
    # data.
    current = 0

    # The data starts with "SV10" followed by 0x04, 0x03, 0x02, 0x01.
    # It's meaningless to me, so after reading it, I discard it.
    size, chunks = _get_chunks(tag, current, "4s4s")
    current += size

    _my_assert(chunks[0] == "SV10")
    _my_assert(chunks[1] == "\4\3\2\1")

    # get the number of elements in the outer list
    size, chunks = _get_chunks(tag, current, "L")
    current += size
    element_count = chunks[0]

    # Eat a delimiter (should be 0x77)
    size, chunks = _get_chunks(tag, current, "4s")
    current += size
    _my_assert(chunks[0] in DELIMITERS)

    for i in range(element_count):
        # Each element looks like this:
        # - (64 bytes) Element name, e.g. ImagedNucleus, NumberOfFrames,
        #   VariableFlipAngleFlag, MrProtocol, etc. Only the data up to the
        #   first 0x00 is important. The rest is helpfully populated with
        #   noise that has enough pattern to make it look like something
        #   other than the garbage that it is.
        # - (4 bytes) VM
        # - (4 bytes) VR
        # - (4 bytes) syngo_dt
        # - (4 bytes) # of subelements in this element (often zero)
        # - (4 bytes) a delimiter (0x4d or 0xcd)
        size, chunks = _get_chunks(tag, current,
                                  "64s" + "4s" + "4s" + "4s" + "L" + "4s")
        current += size

        name, vm, vr, syngo_dt, subelement_count, delimiter = chunks
        _my_assert(delimiter in DELIMITERS)

        # The subelements hold zero or more strings. Those strings are stored
        # temporarily in the values list.
        values = [ ]

        for j in range(subelement_count):
            # Each subelement looks like this:
            # - (4 x 4 = 16 bytes) Call these four bytes A, B, C and D. For
            #   some strange reason, C is always a delimiter, while A, B and
            #   D are always equal to one another. They represent the length
            #   of the associated data string.
            # - (n bytes) String data, the length of which is defined by
            #   A (and A == B == D).
            # - (m bytes) Padding if length is not an even multiple of four.
            size, chunks = _get_chunks(tag, current, "4L")
            current += size

            _my_assert(chunks[0] == chunks[1])
            _my_assert(chunks[1] == chunks[3])
            _my_assert(chunks[2] in DELIMITERS)
            length = chunks[0]

            # get a chunk-o-stuff, length indicated by code above.
            # Note that length can be 0.
            size, chunks = _get_chunks(tag, current, "%ds" % length)
            current += size
            if chunks[0]:
                values.append(chunks[0])

            # If we're not at a 4 byte boundary, move.
            # Clever modulus code below swiped from GDCM
            current += (4 - (length % 4)) % 4

        # The value becomes a single string item (possibly "") or a list
        # of strings
        if len(values) == 0:
            values = ""
        if len(values) == 1:
            values = values[0]

        _my_assert(name not in elements)
        elements[name] = values

    return elements
    



import os
import sys
import math

from nipype.interfaces.traits_extension import isdefined
from nipype.interfaces.base import BaseInterface, InputMultiPath,\
    OutputMultiPath, BaseInterfaceInputSpec, traits, File, TraitedSpec,\
    CommandLineInputSpec, CommandLine, Directory
from traits.trait_errors import TraitError
import nipype.interfaces.dcm2nii as d2n
from duke_siemens.util_dicom_siemens import read as read_siemens_shadow

from util import *
from config import *

class GatherFilesInputSpec(BaseInterfaceInputSpec):
    files = InputMultiPath(
            traits.Either(traits.List(File(exists=True)),File(exists=True)),
            mandatory=True,
            desc="a list of files you would like to gather into one directory with symlinks",
            copyfile=False)

class GatherFilesOutputSpec(TraitedSpec):
    links = OutputMultiPath(
            traits.List(File(exists=True)),
            desc="a list of files, all in the same directory.")

class GatherFiles(BaseInterface):
    input_spec = GatherFilesInputSpec
    output_spec = GatherFilesOutputSpec
    d_files = []
    def _run_interface(self, runtime):
        import os
        files = self.inputs.files
        new_dir = os.path.abspath("link_dir")
        if not os.path.exists(new_dir):
            os.mkdir(new_dir)
        for f in files:
            fname = os.path.split(f)[-1]
            os.link(f, os.path.join(new_dir, fname))
        return runtime

    def _list_outputs(self):
        from glob import glob
        import os
        outputs = self._outputs().get()
        ls = glob(os.path.join(os.path.abspath("."), "link_dir","*"))
        outputs["links"] = ls
        return outputs

class HCDcm2nii(d2n.Dcm2nii):
    """ We override to fix a bug in output listing... """
    def _parse_stdout(self, stdout):
        import re
        import os
        files = []
        reoriented_files = []
        reoriented_and_cropped_files = []
        bvecs = []
        bvals = []
        skip = False
        last_added_file = None
        for line in stdout.split("\n"):
            if not skip:
                file = None
                if line.startswith("Saving "):
                    file = line[len("Saving "):]
                elif line.startswith("GZip..."):
                    #for gzipped outpus files are not absolute
                    if isdefined(self.inputs.output_dir):
                        output_dir = self.inputs.output_dir
                    else:
                        output_dir = self._gen_filename('output_dir')
                    file = os.path.abspath(os.path.join(output_dir,
                                                        line[len("GZip..."):]))
                elif line.startswith("Number of diffusion directions "):
                    if last_added_file:
                        base, filename, ext = split_filename(last_added_file)
                        bvecs.append(os.path.join(base,filename + ".bvec"))
                        bvals.append(os.path.join(base,filename + ".bval"))
                elif re.search('.*->(.*)', line):
                    val = re.search('.*->(.*)', line)
                    val = val.groups()[0]
                    if isdefined(self.inputs.output_dir):
                        output_dir = self.inputs.output_dir
                    else:
                        output_dir = self._gen_filename('output_dir')
                    val = os.path.join(output_dir, val)
                    file = val

                if file:
                    if last_added_file and os.path.exists(file) and not last_added_file in file:
                        files.append(file)
                    last_added_file = file
                    continue

                if line.startswith("Reorienting as "):
                    reoriented_files.append(line[len("Reorienting as "):])
                    skip = True
                    continue
                elif line.startswith("Cropping NIfTI/Analyze image "):
                    base, filename = os.path.split(line[len("Cropping NIfTI/Analyze image "):])
                    filename = "c" + filename
                    reoriented_and_cropped_files.append(os.path.join(base, filename))
                    skip = True
                    continue
            skip = False
        return files, reoriented_files, reoriented_and_cropped_files, bvecs, bvals

class DicomInfoInputSpec(BaseInterfaceInputSpec):
    files = InputMultiPath(
            traits.Either(traits.List(File(exists=True)),File(exists=True)),
            mandatory=True,
            desc="a list of dicom files from which to extract data",
            copyfile=False)

class DicomInfoOutputSpec(TraitedSpec):
    info = traits.List(traits.Dict(),
            desc="an ordered list of dicts, all in the same directory.")

class DicomInfo(BaseInterface):
    input_spec = DicomInfoInputSpec
    output_spec = DicomInfoOutputSpec
    info = []
    def _run_interface(self, runtime):
        import dicom
        files = self.inputs.files
        by_series = {}
        self.info = []
        for f in files:
            d = dicom.read_file(f)
            try: 
                s_num = d.SeriesNumber
                s_num = int(s_num)
            except Exception, e:
                raise e
            if not s_num in by_series:
                by_series[s_num] = {
                    "series_num": s_num,
                    "series_desc": getattr(d,"SeriesDescription",None),
                    "protocol_name": getattr(d,"ProtocolName",None),
                }
                if [0x19,0x1018] in d and \
                        "description" in dir(d[0x19,0x1018]) and \
                        "RealDwellTime" in d[0x19,0x1018].description():
                    try:
                        by_series[s_num]["RealDwellTime"] = float(d[0x19,0x1018].value)
                    except Exception, e:
                        pass # don't die
                it = getattr(d,"ImageType",None)
                if it:
                    if not isinstance(it,str):
                        it = list(it)
                    by_series[s_num]["image_type"] = it
                ipped = getattr(d,"InPlanePhaseEncodingDirection", None)
                if ipped:
                    by_series[s_num]["ipp_encoding_direction"] = ipped
                # ep echo spacing ingredients
                bpppe = d.get((0x0019,0x1028), None)
                if bpppe:
                    try:
                        bpppe = float(bpppe.value)
                    except Exception, e:
                        pass
                    else:
                        by_series[s_num]["bw_per_pix_phase_encode"] = bpppe
                acq_mat = getattr(d, "AcquisitionMatrix", None)
                if acq_mat and len(acq_mat) == 4:
                    # acq mat text gets turned into this struct.
                    # n will be in 0/1 (other index will == 0), m will be in 2/3
                    by_series[s_num]["acq_matrix_n"] = acq_mat[0] or acq_mat[1]
                    by_series[s_num]["acq_matrix_m"] = acq_mat[2] or acq_mat[3]
                # try to get orientation from header
                orient = orientation_from_dcm_header(d)
                if orient:
                    by_series[s_num]["orientation"] = orient
                # try to get siemens shadow header
                try:
                    ss = read_siemens_shadow(f)[0]
                except Exception, e:
                    pass
                else:
                    siemens_keys = ["in_plane_rotation", "polarity_swap"]
                    by_series[s_num].update(dict([(k,v) for k,v in ss.iteritems() if k in siemens_keys]))
            bs = by_series[s_num]
            # collect all of the unique EchoTimes per series
            try:
                et = getattr(d,"EchoTime",None)
                if et:
                    et = float(et)
                    if not "echo_times" in bs:
                        bs["echo_times"] = []
                    if not et in bs["echo_times"]:
                        bs["echo_times"].append(et)
            except Exception, e:
                pass # don't die.
        for s_num, s_info in by_series.iteritems():
            # find delta_te, collapse to echo time if needed
            if "echo_times" in s_info:
                etc = len(s_info["echo_times"])
                if etc == 0:
                    del s_info["echo_times"]
                elif etc == 1:
                    s_info["echo_time"] = s_info["echo_times"][0]
                    del s_info["echo_times"]
                elif etc == 2:
                    s_info["delta_te"] = abs(s_info["echo_times"][0] - s_info["echo_times"][1])
        for k in sorted(by_series.keys()):
            self.info.append(by_series[k])
        return runtime

    def _list_outputs(self):
        outputs = self._outputs().get()
        outputs["info"] = self.info
        return outputs

class NiiWranglerInputSpec(BaseInterfaceInputSpec):
    nii_files = InputMultiPath(
            traits.Either(traits.List(File(exists=True)),File(exists=True)),
            mandatory=True,
            desc="a list of nifti files to be categorized, matched up, etc.",
            copyfile=False)
    series_map = traits.Dict(
            key_trait=traits.Str(),
            value_trait=traits.List(),
            value = {},
            mandatory=False,
            usedefault=True,
            desc="keys are any member of SCAN_TYPES, values are lists of series\
                  descriptions as recorded in DICOM headers.")
    dicom_info = traits.List(
            mandatory=True,
            desc="one dict for each series in the session, in the order they were\
                  run. each dict should contain at least the series_num (int) and\
                  the series_desc (str).")
    ep_echo_spacing = traits.Either(traits.Enum("NONE"), traits.Float(),
            desc="""
            The effective echo spacing of your BOLD images. Already accounts
            for whether or not iPAT (acceleration in the phase direction) was
            used. If you're using acceleration, then the EES is not going to
            match the 'Echo Spacing' that Siemen's reports in the console.

            Setting this value will prevent any attempt to derive it.""")
    ep_unwarp_dir = traits.Enum("x", "x-", "-x", "y", "y-", "-y", "z", "z-", "-z",
            desc="Setting this value will prevent any attempt to derive it.")
    block_struct_averaging = traits.Bool(False,
            mandatory=False, usedefault=True,
            desc="""
            Causes us to only use the first t1 and t2 images. A kludge for
            some data that fails during structural averaging.""")
    ep_fieldmap_selection = traits.Enum("first","most_recent",
            mandatory=False, usedefault=True,
            desc="""
            If you have more than one set of ep fieldmaps, then you can either
            use the first set during preprocessing of all bold images
            ('first'), or you can select the se fieldmap set that was taken
            most recently prior to acquisition of a given bold image, or -
            failing that - the first available se fieldmap thereafter
            ('most_recent')."""
            )

class NiiWranglerOutputSpec(TraitedSpec):
    t1_structs = OutputMultiPath(
            traits.List(File(exists=True)),
            mandatory=True,
            desc="a list of t1 niftis in chronological order.")
    t2_structs = OutputMultiPath(
            traits.List(File(exists=True)),
            mandatory=True,
            desc="a list of t2 niftis in chronological order.")
    bolds = OutputMultiPath(
            traits.List(File(exists=True)),
            mandatory=True,
            desc="a list of BOLD niftis in chronological order.")
    bold_names = traits.List(traits.Str(),
            mandatory=True,
            desc="a list of names for the bold images. Length must match number of bold images.")
    sb_refs = OutputMultiPath(
            traits.List(File(exists=True)),
            mandatory=True,
            desc="a list of BOLD_SBRef niftis in chronological order. Length must match number of bold images.")
    pos_fieldmaps = OutputMultiPath(
            traits.List(File(exists=True)),
            mandatory=True,
            desc="a list of positive fieldmap niftis. Length must match number of bold images.")
    neg_fieldmaps = OutputMultiPath(
            traits.List(File(exists=True)),
            mandatory=True,
            desc="a list of negative fieldmap niftis. Length must match number of bold images.")
    mag_fieldmap = traits.Either(traits.Enum("NONE"), traits.File(exists=True),
            default="NONE", mandatory=False, usedefault=True,
            desc="magnitude fieldmap image for structural images. first matching image will be used.")
    phase_fieldmap = traits.Either(traits.Enum("NONE"), traits.File(exists=True),
            default="NONE", mandatory=False, usedefault=True,
            desc="phase fieldmap image for structural images. first matching image will be used.")
    t1_sample_spacing = traits.Either(traits.Float(), traits.Enum("NONE"),
            mandatory=False, usedefault=True,
            desc="DICOM field (0019,1018) * 10e-9 in s or 'NONE' if not used. (float) seconds.")
    t2_sample_spacing = traits.Either(traits.Float(), traits.Enum("NONE"),
            mandatory=False, usedefault=True,
            desc="DICOM field (0019,1018) * 10e-9 in s or 'NONE' if not used. (float) seconds.")
    fieldmap_te = traits.Either(traits.Enum("NONE"), traits.Float(),
            default="NONE", mandatory=False, usedefault=True,
            desc="delta TE in ms for magnitude fieldmap or 'NONE' if not used.")
    ep_echo_spacings = traits.List(traits.Either(traits.Enum("NONE"), traits.Float()),
            value=["NONE"], mandatory=False, usedefault=True,
            desc="""
            The effective echo spacing of your BOLD images. Already accounts
            for whether or not iPAT (acceleration in the phase direction) was
            used. If you're using acceleration, then the EES is not going to
            match the 'Echo Spacing' that Siemen's reports in the console.

            This value will be derived, if not overridden by the input of the
            same name. Please inspect the value after your initial run of the
            pipeline to ensure that it's sane.

            Length must match number of bold images.""")
    ep_unwarp_dirs = traits.List(traits.Enum("x", "x-", "-x", "y", "y-", "-y", "z", "z-", "-z",),
            mandatory=True,
            desc="Length must match number of bold images.")

class NiiWrangler(BaseInterface):
    input_spec = NiiWranglerInputSpec
    output_spec = NiiWranglerOutputSpec
    t1_files = []
    t2_files = []
    bolds = []
    bold_names = []
    sbrefs = []
    fieldmap_pos = []
    fieldmap_neg = []
    fieldmap_mag = []
    fieldmap_ph = []
    fieldmap_mag_delta_te = "NONE"
    t1_sample_spacing = 0.
    t2_sample_spacing = 0.
    ep_e_spaces = None
    ep_unwarp_dirs = None
    def _run_interface(self, runtime):
        import re
        import operator
        nii_files = self.inputs.nii_files
        smap = self.inputs.series_map
        dinfo = self.inputs.dicom_info
        block_averaging = self.inputs.block_struct_averaging
        s_num_reg = re.compile(".*s(\d+)a(?!.*/)") # sux to use filename. make more robust if needed.
        nii_by_series = {}
        fails = []
        extras = []
        for fn in nii_files:
            try:
                # we only want the first nii for each series
                # TODO: find out what those others (A/B) are all about. fix this as needed.
                sn = int(s_num_reg.match(fn).groups()[0])
                if sn in nii_by_series:
                    extras.append(fn)
                    continue
                nii_by_series[sn] = fn
            except Exception, e:
                fails.append(fn)
        if fails:
            raise ValueError("Could not derive series number from file names: %s." % str(fails))
        if extras:
            print >> sys.stderr, "\nWARNING: Ignoring extra niftis: %s\n" % str(extras)
        # add nifti names to the dicts
        m_count = 0
        for sn, fn in nii_by_series.iteritems():
            m = filter(lambda x: x.get("series_num",-1) == sn, dinfo)
            if not m:
                continue
            m_count += 1
            m[0]["nifti_file"] = fn
        if not m_count == len(dinfo):
            raise ValueError("incorrect number of nifti->series matches (%d/%d)" % (m_count, len(dinfo)))
        # time for some data wrangling
        nf = "nifti_file"
        sd = "series_desc"
        it = "image_type"
        t1fs = [d for d in filter(lambda x: sd in x and x[sd] in smap.get("t1",[]), dinfo) if nf in d]
        t2fs = [d for d in filter(lambda x: sd in x and x[sd] in smap.get("t2",[]), dinfo) if nf in d]
        if block_averaging:
            t1fs = [t1fs[0]]
            t2fs = [t2fs[0]]
        self.t1_files = [d[nf] for d in t1fs]
        self.t2_files = [d[nf] for d in t2fs]
        bs = [d for d in filter(lambda x: sd in x and x[sd] in smap.get("bold",[]), dinfo) if nf in d]
        self.bolds = [d[nf] for d in bs]
        self.sbrefs = [d[nf] for d in filter(lambda x: sd in x and x[sd] in smap.get("bold_sbref",[]), dinfo) if nf in d]
        # for now, we only support one se fieldmap pair. In the future, we'll implement
        # a more flexible policy here. This is actually really crappy... but the user can always
        # flip the unwarpdir through config :/
        s_policy = self.inputs.ep_fieldmap_selection
        pos_types = reduce(operator.add, [smap.get(k, []) for k in POS_FIELDMAPS])
        neg_types = reduce(operator.add, [smap.get(k, []) for k in NEG_FIELDMAPS])
        pos = [d for d in filter(lambda x: sd in x and x[sd] in pos_types, dinfo) if nf in d]
        neg = [d for d in filter(lambda x: sd in x and x[sd] in neg_types, dinfo) if nf in d]
        both = zip(pos,neg)
        if s_policy == "most_recent":
            pfs = []
            nfs = []
            for bold in bs:
                sn = bold["series_num"]
                # we want the last of the images with a lower sn, or the firs tof the images with a higher sn
                earlier = filter(lambda x: x[0]["series_num"] < sn and x[1]["series_num"] < sn, both)
                later = filter(lambda x: x[0]["series_num"] < sn and x[1]["series_num"] < sn, both)
                if earlier:
                    pfs.append(earlier[-1][0][nf])
                    nfs.append(earlier[-1][1][nf])
                elif later:
                    pfs.append(later[0][0][nf])
                    nfs.append(later[0][1][nf])
                else:
                    print "This... should never happen."
            self.fieldmaps_pos = pfs
            self.fieldmaps_neg = nfs
        else: # default to "first"
            self.fieldmaps_pos = [pos[0][nf] for n in self.bolds] if pos else []
            self.fieldmaps_neg = [neg[0][nf] for n in self.bolds] if pos else []
        # we have to do some extra looking at the headers for the mag and phase fieldmaps too
        mag_fs = filter(lambda x: sd in x and
                x[sd] in smap.get("fieldmap_magnitude",[]) and
                it in x and
                isinstance(x[it], list) and
                len(x[it]) > 2 and
                x[it][2].strip().lower() == "m", dinfo) # we want the 3rd field of image type to be 'm'
        phase_fs = filter(lambda x: sd in x and
                x[sd] in smap.get("fieldmap_phase",[]) and
                it in x and
                isinstance(x[it], list) and
                len(x[it]) > 2 and
                x[it][2].strip().lower() == "p", dinfo) # we want the 3rd field of image type to be 'p'
        self.fieldmap_mag = [d[nf] for d in mag_fs if nf in d]
        self.fieldmap_ph = [d[nf] for d in phase_fs if nf in d]
        # calculate echo spacing for each of the bold images
        ep_calc_fail = False
        if isdefined(self.inputs.ep_echo_spacing):
            self.ep_e_spaces = [self.inputs.ep_echo_spacing for n in self.bolds]
        elif bs and any(["bw_per_pix_phase_encode" in d and "acq_matrix_n" in d for d in bs]):
            if not all(["bw_per_pix_phase_encode" in d and "acq_matrix_n" in d for d in bs]):
                ep_calc_fail = True
            else:
                self.ep_e_spaces = [1/(d["bw_per_pix_phase_encode"] * d["acq_matrix_n"]) for d in bs]
        else:
            self.ep_e_spaces = ["NONE" for n in self.bolds]
        # output delta te for magnitude fieldmap if available
        if mag_fs and self.fieldmap_mag and self.fieldmap_ph:
            self.fieldmap_mag_delta_te = mag_fs[0].get("delta_te","NONE")
        else:
            self.fieldmap_mag_delta_te = "NONE"
        # a BOLD image by any other name...
        self.bold_names = ["bold_%d" % n for n in xrange(len(self.bolds))]
        # we'll derive the ep unwarp dir in the future... for now, just set it in config
        if isdefined(self.inputs.ep_unwarp_dir):
            self.ep_unwarp_dirs = [self.inputs.ep_unwarp_dir for n in self.bolds]
            # if you have any polarity swapped series, apply that now
            pswaps = smap.get("polarity_swapped", [])
            if pswaps:
                for b_idx, uw_dir in enumerate(self.ep_unwarp_dirs):
                    if bs[b_idx].get("series_desc",None) in pswaps:
                        raw_dir = uw_dir.replace("-","")
                        self.ep_unwarp_dirs[b_idx] = "-"+raw_dir if not "-" in uw_dir else raw_dir
        else:
            # fail. we can do better!
            raise ValueError("We can't derive ep_unwarp_dir yet. Please set it in the nii wrangler config section.")
        ### and let's do some sanity checking
        # warn if you're going to ignore field or magnitude phase maps
        if not (len(self.fieldmap_mag) and len(self.fieldmap_ph)):
            print >> sys.stderr, "\nWARNING: found %d magnitude fieldmaps and %d phase fieldmaps.\n" % (len(self.fieldmap_mag), len(self.fieldmap_ph))
        # if we have any sbrefs, we should have one for each bold
        if self.sbrefs and len(self.sbrefs) != len(self.bolds):
            raise ValueError("Had %d bolds, but %d SBRefs. If there are any SBRefs, there must be one for each BOLD image"
                             % (len(self.bolds), len(self.sbrefs)))
        # and if we DON'T have any, then we need a list full of NONE
        if not self.sbrefs:
            self.sbrefs = ["NONE" for n in self.bolds]
        # we should have the same number of positive and negative fieldmaps
        if len(self.fieldmaps_pos) != len(self.fieldmaps_neg):
            raise ValueError("Mismatched number of pos and neg fieldmaps (%d/%d)" % (len(self.fieldmaps_pos), len(self.fieldmaps_neg)))
        # we also need the same number of pos/neg se fieldmaps as bold images
        if len(self.fieldmaps_pos) != len(self.bolds):
            raise ValueError("Mismatched number of pos/neg fieldmaps and bolds (%d/%d)" % (len(self.fieldmaps_pos), len(self.bolds)))
        ### derive the derived values
        # t1_sample_spacing
        if t1fs and "RealDwellTime" in t1fs[0].keys():
            self.t1_sample_spacing = t1fs[0]["RealDwellTime"] * math.pow(10,-9)
        else:
            self.t1_sample_spacing = "NONE"
        # t2_sample_spacing
        if t2fs and "RealDwellTime" in t2fs[0].keys():
            self.t2_sample_spacing = t2fs[0]["RealDwellTime"] * math.pow(10,-9)
        else:
            self.t2_sample_spacing = "NONE"
        # don't continue if there was screwiness around calculating ep echo spacing
        if ep_calc_fail:
            raise ValueError("Unabel to calculate ep echo spacing. Try specifying manually in nii wrangler config section.")
        return runtime

    def _list_outputs(self):
        outputs = self._outputs().get()
        outputs["t1_structs"] = self.t1_files
        outputs["t2_structs"] = self.t2_files
        outputs["bolds"] = self.bolds
        outputs["bold_names"] = self.bold_names
        outputs["sb_refs"] = self.sbrefs
        outputs["pos_fieldmaps"] = self.fieldmaps_pos
        outputs["neg_fieldmaps"] = self.fieldmaps_neg
        fm = self.fieldmap_mag
        fp = self.fieldmap_ph
        outputs["mag_fieldmap"] = fm[0] if fm and fp else "NONE"
        outputs["phase_fieldmap"] = fp[0] if fm and fp else "NONE"
        if not (fm and fp):
            print >> sys.stderr, "\nWARNING: not using magnitude or phase fieldmaps.\n"
        outputs["t1_sample_spacing"] = self.t1_sample_spacing
        outputs["t2_sample_spacing"] = self.t2_sample_spacing
        outputs["fieldmap_te"] = self.fieldmap_mag_delta_te
        outputs["ep_echo_spacings"] = self.ep_e_spaces
        outputs["ep_unwarp_dirs"] = self.ep_unwarp_dirs
        return outputs

class HCPCommandInputSpec(CommandLineInputSpec):
    full_command = traits.Str("PreFreeSurferPipeline.sh",
            mandatory=False,
            usedefault=True,
            desc="Full path to the relevant HCP script.")

class HCPCommandOutputSpec(TraitedSpec):
    subject = traits.Str(
            mandatory=True,
            desc="subject id or number, as a string",)

class HCPCommand(CommandLine):
    """ Handles setting of _cmd to full path.

    Input spec must be a subclass of HCPCommandInputSpec.
    """
    def _run_interface(self, runtime):
        import os
        self._cmd = self.inputs.full_command
        return super(HCPCommand, self)._run_interface(runtime)

class PreFSInputSpec(HCPCommandInputSpec):
    full_command = traits.Str("PreFreeSurferPipeline.sh",
            mandatory=False,
            usedefault=True,
            desc="Full path to the relevant HCP script.")
    study_dir = traits.Either(traits.Enum(None), Directory(exists=True),
            default=None, mandatory=False, usedefault=True, position=0, argstr='--path="%s"',
            desc="Leave as None to have the interface make a new dir in its working dir.",)
    subject = traits.Str(
            mandatory=True, position=1, argstr='--subject="%s"',
            desc="",)
    t1_files = InputMultiPath(
            traits.Either(traits.List(File(exists=True)),File(exists=True)),
            mandatory=True, sep="@", position=2, argstr='--t1="%s"',
            desc="t1 files.")
    t2_files = InputMultiPath(
            traits.Either(traits.List(File(exists=True)),File(exists=True)),
            mandatory=True, sep="@", position=3, argstr='--t2="%s"',
            desc="t2 files.")
    t1_template = InputMultiPath(File(exists=True),
            mandatory=True, position=4, argstr='--t1template="%s"',
            desc="MNI0.7mm template")
    t1_template_brain = InputMultiPath(File(exists=True),
            mandatory=True, position=5, argstr='--t1templatebrain="%s"',
            desc="Brain extracted MNI0.7mm template")
    t1_template_2mm = InputMultiPath(File(exists=True),
            mandatory=True, position=6, argstr='--t1template2mm="%s"',
            desc="MNI2mm template")
    t2_template = InputMultiPath(File(exists=True),
            mandatory=True, position=7, argstr='--t2template="%s"',
            desc="MNI0.7mm T2wTemplate")
    t2_template_brain = InputMultiPath(File(exists=True),
            mandatory=True, position=8, argstr='--t2templatebrain="%s"',
            desc="Brain extracted MNI0.7mm T2wTemplate")
    t2_template_2mm = InputMultiPath(File(exists=True),
            mandatory=True, position=9, argstr='--t2template2mm="%s"',
            desc="MNI2mm T2wTemplate")
    template_mask = InputMultiPath(File(exists=True),
            mandatory=True, position=10, argstr='--templatemask="%s"',
            desc="Brain mask MNI0.7mm template")
    template_2mm_mask = InputMultiPath(File(exists=True),
            mandatory=True, position=11, argstr='--template2mmmask="%s"',
            desc="MNI2mm template")
    brain_size = traits.Int(150,
            mandatory=False, usedefault=True, position=12, argstr='--brainsize="%d"',
            desc="brain size in mm")
    fnirt_config = InputMultiPath(File(exists=True),
            mandatory=True, position=13, argstr='--fnirtconfig="%s"',
            desc="FNIRT 2mm T1w Config")
    fieldmap_magnitude = traits.Either(traits.Enum("NONE"), traits.File(exists=True),
            default="NONE", mandatory=False, usedefault=True, position=14, argstr='--fmapmag="%s"',
            xor=["se_fieldmap_neg","se_fieldmap_pos","se_echo_spacing","se_unwarp_dir"],
            desc="Expects 4D magitude volume with two 3D timepoints or 'NONE' if not used")
    fieldmap_phase = traits.Either(traits.Enum("NONE"), traits.File(exists=True),
            default="NONE", mandatory=False, usedefault=True, position=15, argstr='--fmapphase="%s"',
            xor=["se_fieldmap_neg","se_fieldmap_pos","se_echo_spacing","se_unwarp_dir"],
            desc="Expects 3D phase difference volume or 'NONE' if not used")
    fieldmap_te = traits.Either(traits.Enum("NONE"), traits.Float(),
            default="NONE", mandatory=False, usedefault=True, position=16, argstr='--echodiff="%s"',
            xor=["se_fieldmap_neg","se_fieldmap_pos","se_echo_spacing","se_unwarp_dir"],
            desc="delta TE in ms for magnitude field map or 'NONE' if not used.")
    se_fieldmap_neg = traits.Either(traits.Enum("NONE"), traits.File(exists=True),
            default="NONE", mandatory=False, usedefault=True, position=17, argstr='--SEPhaseNeg="%s"',
            xor=["fieldmap_magnitude", "fieldmap_phase", "fieldmap_te"],
            desc="""
            For the spin echo field map volume with a negative phase encoding
            direction (LR in HCP data), set to NONE if using regular
            FIELDMAP""")
    se_fieldmap_pos = traits.Either(traits.Enum("NONE"), traits.File(exists=True),
            default="NONE", mandatory=False, usedefault=True, position=18, argstr='--SEPhasePos="%s"',
            xor=["fieldmap_magnitude", "fieldmap_phase", "fieldmap_te"],
            desc="""
            For the spin echo field map volume with a positive phase encoding
            direction (RL in HCP data), set to NONE if using regular
            FIELDMAP""")
    se_echo_spacing = traits.Either(traits.Enum("NONE"), traits.Float(),
            default="NONE", mandatory=False, usedefault=True, position=19, argstr='--echospacing="%s"',
            xor=["fieldmap_magnitude", "fieldmap_phase", "fieldmap_te"],
            desc="Echo Spacing or Dwelltime of Spin Echo Field Map or 'NONE' if not used")
    se_unwarp_dir = traits.Enum("NONE", "x", "y",
            defaut="NONE", mandatory=False, usedefault=True, position=20, argstr='--seunwarpdir="%s"',
            xor=["fieldmap_magnitude", "fieldmap_phase", "fieldmap_te"],
            desc="x or y (minus or not does not matter) 'NONE' if not used")
    t1_sample_spacing = traits.Either(traits.Enum("NONE"), traits.Float(),
            default="NONE", mandatory=False, usedefault=True, position=21, argstr='--t1samplespacing="%s"',
            desc="DICOM field (0019,1018) * 10e-9 in s or 'NONE' if not used.")
    t2_sample_spacing = traits.Either(traits.Enum("NONE"), traits.Float(),
            default="NONE", mandatory=False, usedefault=True, position=22, argstr='--t2samplespacing="%s"',
            desc="DICOM field (0019,1018) * 10e-9 in s or 'NONE' if not used.")
    unwarp_dir = traits.Enum("z", "z-", "-z", "x", "x-", "-x", "y", "y-", "-y", "NONE",
            mandatory=False, usedefault=True, position=23, argstr='--unwarpdir="%s"',
            desc="z appears to be best or 'NONE' if not used")
    grad_distort_coeffs = traits.Either(traits.Enum("NONE"), traits.File(exists=True),
            default="NONE", mandatory=False, usedefault=True, position=24, argstr='--gdcoeffs="%s"',
            desc="Location of Coeffs file or 'NONE' to skip")
    avg_rcd_method = traits.Enum("FIELDMAP","TOPUP","NONE",
            mandatory=False, usedefault=True, position=25, argstr='--avgrdcmethod="%s"',
            desc="""
    Averaging and readout distortion correction methods: "NONE" = average any
    repeats with no readout correction "FIELDMAP" = average any repeats and use
    field map for readout correction "TOPUP" = average and distortion correct at
    the same time with topup/applytopup only works for 2 images currently""")
    top_up_config = traits.Either(traits.Enum("NONE"), traits.File(exists=True),
            default="NONE", mandatory=False, usedefault=True, position=26, argstr='--topupconfig="%s"',
            desc="Config for topup or 'NONE' if not used")
    print_com = traits.Enum("","echo",
            mandatory=False, usedefault=True, position=27, argstr='--printcom=%s',
            desc="")

class PreFSOutputSpec(HCPCommandOutputSpec):
    study_dir = Directory(
            mandatory=True,
            exists=True,
            desc='the study dir')
    subject_dir = Directory(
            mandatory=False,
            exists=True,
            desc='the subject dir')
    subject_t1_dir = Directory(
            mandatory=False,
            exists=True,
            desc='the subject dir')
    t1_acpc_dc_restore = traits.File(
            mandatory=True,
            exists=True,
            desc="The resulting T1w_acpc_dc_restore.nii.gz")
    t1_acpc_dc_restore_brain = traits.File(
            mandatory=True,
            exists=True,
            desc="The resulting T1w_acpc_dc_restore_brain.nii.gz")
    t2_acpc_dc_restore = traits.File(
            mandatory=True,
            exists=True,
            desc="The resulting T2w_acpc_dc_restore.nii.gz")

class PreFS(HCPCommand):
    input_spec = PreFSInputSpec
    output_spec = PreFSOutputSpec
    _cmd = "PreFreeSurferPipeline.sh"
    
    def _format_arg(self, name, spec, value):
        if name not in ["t1_sample_spacing", "t2_sample_spacing", "fieldmap_te"]:
            return super(PreFS, self)._format_arg(name, spec, value)
        st = value
        if isinstance(st, float):
            st = "%.9f" % st if name in ["t1_sample_spacing", "t2_sample_spacing"] else str(st)
        st = str(st) if not isinstance(st, str) else st
        return spec.argstr%st

    def _run_interface(self, runtime):
        import os
        sf = self.inputs.study_dir
        if not sf:
            sf = "./study_dir"
        sf = os.path.abspath(sf)
        if not os.path.exists(sf):
            os.mkdir(sf)
        self.inputs.study_dir = sf
        return super(PreFS, self)._run_interface(runtime)

    def _list_outputs(self):
        outputs = self._outputs().get()
        outputs["study_dir"] = self.inputs.study_dir
        outputs["subject"] = self.inputs.subject
        subj_d = os.path.join(self.inputs.study_dir, self.inputs.subject)
        if os.path.exists(subj_d):
            outputs["subject_dir"] = subj_d
        subj_t1_d = os.path.join(subj_d, "T1w")
        if os.path.exists(subj_t1_d):
            outputs["subject_t1_dir"] = subj_t1_d
            files = {"T1w_acpc_dc_restore.nii.gz":"t1_acpc_dc_restore",
                     "T1w_acpc_dc_restore_brain.nii.gz":"t1_acpc_dc_restore_brain",
                     "T2w_acpc_dc_restore.nii.gz":"t2_acpc_dc_restore"}
            for f_name, v_name in files.iteritems():
                p = os.path.join(subj_t1_d, f_name)
                if not os.path.exists(p):
                    continue
                outputs[v_name] = p
        return outputs

class FSInputSpec(HCPCommandInputSpec):
    full_command = traits.Str("FreeSurferPipeline.sh",
            mandatory=True,
            desc="Full path to the relevant HCP script.")
    subject = traits.Str(
            mandatory=True, position=0, argstr='--subject="%s"',
            desc="",)
    subject_t1_dir = InputMultiPath(Directory(exists=True),
            mandatory=True, position=1, argstr='--subjectDIR="%s"',
            desc="path to T1w dir created by the pre_freesurfer node.")
    t1_acpc_dc_restore = InputMultiPath(File(exists=True),
            mandatory=True, position=2, argstr='--t1="%s"',
            desc="")
    t1_acpc_dc_restore_brain = InputMultiPath(File(exists=True),
            mandatory=True, position=3, argstr='--t1brain="%s"',
            desc="")
    t2_acpc_dc_restore = InputMultiPath(File(exists=True),
            mandatory=True, position=4, argstr='--t2="%s"',
            desc="")
    print_com = traits.Enum("","echo",
            mandatory=False, usedefault=True, position=5, argstr='--printcom=%s',
            desc="")

class FSOutputSpec(HCPCommandOutputSpec):
    pass

class FS(HCPCommand):
    input_spec = FSInputSpec
    output_spec = FSOutputSpec
    _cmd = "FreeSurferPipeline.sh"
    
    def _list_outputs(self):
        outputs = self._outputs().get()
        outputs["subject"] = self.inputs.subject
        return outputs

class PostFSInputSpec(HCPCommandInputSpec):
    full_command = traits.Str("PostFreeSurferPipeline.sh",
            mandatory=True,
            desc="Full path to the relevant HCP script.")
    study_dir = Directory(
            mandatory=True, exists=True, position=0, argstr='--path="%s"',
            desc="The study dir. Should contain a subdir named after the subject.",)
    subject = traits.Str(
            mandatory=True, position=1, argstr='--subject="%s"',
            desc="",)
    surf_atlas_dir = Directory(
            mandatory=True, exists=True, position=2, argstr='--surfatlasdir="%s"',
            desc="",)
    grayordinates_dir = Directory(
            mandatory=True, exists=True, position=3, argstr='--grayordinatesdir="%s"',
            desc="",)
    grayordinates_res = traits.Int(2,
            mandatory=False, usedefault=True, position=4, argstr='--grayordinatesres="%d"',
            desc="""
            Usually 2mm.""")
    high_res_mesh = traits.Int(164,
            mandatory=False, usedefault=True, position=5, argstr='--hiresmesh="%d"',
            desc="Usually 164k vertices",)
    low_res_mesh = traits.Int(32,
            mandatory=False, usedefault=True, position=6, argstr='--lowresmesh="%d"',
            desc="Usually 32k vertices",)
    subcort_gray_labels = InputMultiPath(File(exists=True),
            mandatory=True, position=7, argstr='--subcortgraylabels="%s"',
            desc="")
    freesurfer_labels = InputMultiPath(File(exists=True),
            mandatory=True, position=8, argstr='--freesurferlabels="%s"',
            desc="")
    ref_myelin_maps = InputMultiPath(File(exists=True),
            mandatory=True, position=9, argstr='--refmyelinmaps="%s"',
            desc="")
    reg_name = traits.Enum("FS", "MSMSulc",
            default="FS", mandatory=False, usedefault=True, position=10, argstr='--regname=%s',
            desc="MSMSulc is recommended, if binary is not available use FS (FreeSurfer)")
    print_com = traits.Enum("","echo",
            mandatory=False, usedefault=True, position=11, argstr='--printcom=%s',
            desc="")

class PostFSOutputSpec(HCPCommandOutputSpec):
    grayordinates_res = traits.Int(desc="Usually 2mm.")
    high_res_mesh = traits.Int(desc="Usually 164k vertices",)
    low_res_mesh = traits.Int(desc="Usually 32k vertices",)

class PostFS(HCPCommand):
    input_spec = PostFSInputSpec
    output_spec = PostFSOutputSpec
    _cmd = "PostFreeSurferPipeline.sh"

    def _list_outputs(self):
        outputs = self._outputs().get()
        outputs["subject"] = self.inputs.subject
        outputs["grayordinates_res"] = self.inputs.grayordinates_res
        outputs["high_res_mesh"] = self.inputs.high_res_mesh
        outputs["low_res_mesh"] = self.inputs.low_res_mesh
        return outputs

class VolumeProcessingInputSpec(HCPCommandInputSpec):
    full_command = traits.Str("PostFreeSurferPipeline.sh",
            mandatory=True,
            desc="Full path to the relevant HCP script.")
    study_dir = Directory(
            mandatory=True, exists=True, position=0, argstr='--path="%s"',
            desc="The study dir. Should contain a subdir named after the subject.",)
    subject = traits.Str(
            mandatory=True, position=1, argstr='--subject="%s"',
            desc="",)
    bold_name = traits.Str(
            mandatory=True, position=2, argstr='--fmriname="%s"',
            desc="For some reason, we need a unique name for each BOLD series. Whatever. Provide.")
    bold_img = InputMultiPath(File(exists=True),
            mandatory=True, position=3, argstr='--fmritcs="%s"',
            desc="A 4D BOLD image.")
    bold_scout = traits.Either(traits.Enum("NONE"), traits.File(exists=True),
            mandatory=True, position=4, argstr='--fmriscout="%s"',
            desc="""
            A single band reference image (SBRef) is recommended if using
            multiband, set to NONE if you want to use the first volume of the
            timeseries for motion correction""")
    se_fieldmap_neg = traits.Either(traits.Enum("NONE"), traits.File(exists=True),
            default="NONE", mandatory=False, usedefault=True, position=5, argstr='--SEPhaseNeg="%s"',
            # xor=["fieldmap_mag", "fieldmap_phase", "fieldmap_delta_te"],
            requires=["se_fieldmap_pos", "fieldmap_echo_spacing"],
            desc="")
    se_fieldmap_pos = traits.Either(traits.Enum("NONE"), traits.File(exists=True),
            default="NONE", mandatory=False, usedefault=True, position=6, argstr='--SEPhasePos="%s"',
            # xor=["fieldmap_mag", "fieldmap_phase", "fieldmap_delta_te"],
            requires=["se_fieldmap_neg", "fieldmap_echo_spacing"],
            desc="")
    fieldmap_mag = traits.Either(traits.Enum("NONE"), traits.File(exists=True),
            default="NONE", mandatory=False, usedefault=True, position=7, argstr='--fmapmag="%s"',
            # xor=["se_fieldmap_neg", "se_fieldmap_pos", "fieldmap_echo_spacing"],
            requires=["fieldmap_phase", "fieldmap_delta_te"],
            desc="")
    fieldmap_phase = traits.Either(traits.Enum("NONE"), traits.File(exists=True),
            default="NONE", mandatory=False, usedefault=True, position=8, argstr='--fmapphase="%s"',
            # xor=["se_fieldmap_neg", "se_fieldmap_pos", "fieldmap_echo_spacing"],
            requires=["fieldmap_mag", "fieldmap_delta_te"],
            desc="")
    fieldmap_echo_spacing = traits.Either(traits.Enum("NONE"), traits.Float(),
            default="NONE", mandatory=False, usedefault=True, position=9, argstr='--echospacing="%s"',
            # xor=["fieldmap_mag", "fieldmap_phase", "fieldmap_delta_te"],
            requires=["se_fieldmap_neg", "se_fieldmap_pos"],
            desc="(seconds) Spacing or Dwelltime of fMRI image.")
    fieldmap_delta_te = traits.Either(traits.Enum("NONE"), traits.Float(),
            default="NONE", mandatory=False, usedefault=True, position=10, argstr='--echodiff="%s"',
            # xor=["se_fieldmap_neg", "se_fieldmap_pos", "fieldmap_echo_spacing"],
            requires=["fieldmap_mag", "fieldmap_phase"],
            desc="(ms) 2.46ms for 3T, 1.02ms for 7T, set to NONE if using TOPUP")
    unwarp_dir = traits.Enum("x", "x-", "-x", "y", "y-", "-y", "z", "z-", "-z",
            mandatory=False, usedefault=True, position=11, argstr='--unwarpdir="%s"',
            desc="")
    fmri_res = traits.Int(2,
            mandatory=False, usedefault=True, position=12, argstr='--fmrires="%d"',
            desc="Needs to match what is in PostFS. Target final resolution of fMRI data. 2mm is recommended.",)
    distortion_correction_method = traits.Enum("TOPUP", "FIELDMAP",
            mandatory=False, usedefault=True, position=13, argstr='--dcmethod="%s"',
            desc="FIELDMAP or TOPUP, distortion correction is required for accurate processing")
    grad_distort_coeffs = traits.Either(traits.Enum("NONE"), traits.File(exists=True),
            default="NONE", mandatory=False, usedefault=True, position=14, argstr='--gdcoeffs="%s"',
            desc="Gradient distortion correction coefficents, set to NONE to turn off.")
    top_up_config = traits.Either(traits.Enum("NONE"), traits.File(exists=True),
            default="NONE", mandatory=False, usedefault=True, position=15, argstr='--topupconfig="%s"',
            desc="Topup config if using TOPUP, set to NONE if using FIELDMAP as distortion_correction_method.")
    print_com = traits.Enum("","echo",
            mandatory=False, usedefault=True, position=16, argstr='--printcom=%s',
            desc="")

class VolumeProcessingOutputSpec(HCPCommandOutputSpec):
    bold_name = traits.Str(
            mandatory=True,
            desc="A unique name for the BOLD series. We only handle one at a time, so use a mapnode if you have more.")

class VolumeProcessing(HCPCommand):
    input_spec = VolumeProcessingInputSpec
    output_spec = VolumeProcessingOutputSpec
    _cmd = "GenericfMRIVolumeProcessingPipeline.sh"

    def _format_arg(self, name, spec, value):
        if name in ["fieldmap_echo_spacing", "fieldmap_delta_te"]:
            st = value
            if isinstance(st, float):
                st = "%.7f" % st if name in ["fieldmap_echo_spacing",] else str(st)
            st = str(st) if not isinstance(st, str) else st
            return spec.argstr%st
        return super(VolumeProcessing, self)._format_arg(name, spec, value)

    def _list_outputs(self):
        outputs = self._outputs().get()
        outputs["subject"] = self.inputs.subject
        outputs["bold_name"] = self.inputs.bold_name
        return outputs

class SurfaceProcessingInputSpec(HCPCommandInputSpec):
    full_command = traits.Str("GenericfMRISurfaceProcessingPipeline.sh",
            mandatory=True,
            desc="Full path to the relevant HCP script.")
    study_dir = Directory(
            mandatory=True, exists=True, position=0, argstr='--path="%s"',
            desc="The study dir. Should contain a subdir named after the subject.",)
    subject = traits.Str(
            mandatory=True, position=1, argstr='--subject="%s"',
            desc="",)
    bold_name = traits.Str(
            mandatory=True, position=2, argstr='--fmriname="%s"',
            desc="For some reason, we need a unique name for each BOLD series. Whatever. Provide.")
    low_res_mesh = traits.Int(32,
            mandatory=False, usedefault=True, position=3, argstr='--lowresmesh="%d"',
            desc="Needs to match what is in PostFS. Usually 32k vertices.")
    fmri_res = traits.Int(2,
            mandatory=False, usedefault=True, position=4, argstr='--fmrires="%d"',
            desc="Needs to match what is in PostFS. Target final resolution of fMRI data. 2mm is recommended.")
    smoothing_fwhm = traits.Int(2,
            mandatory=False, usedefault=True, position=5, argstr='--smoothingFWHM="%d"',
            desc="Recommended to be roughly the voxel size.")
    grayordinates_res = traits.Int(2,
            mandatory=False, usedefault=True, position=6, argstr='--grayordinatesres="%d"',
            desc="""
            Needs to match what is in PostFS. Usually 2mm. Could be the same
            as FinalfRMIResolution or something different, which will call a
            different module for subcortical processing.""")

class SurfaceProcessingOutputSpec(HCPCommandOutputSpec):
    study_dir = Directory(
            desc="The study dir. Should contain a subdir named after the subject.")

class SurfaceProcessing(HCPCommand):
    input_spec = SurfaceProcessingInputSpec
    output_spec = SurfaceProcessingOutputSpec
    _cmd = "GenericfMRISurfaceProcessingPipeline.sh"

    def _list_outputs(self):
        outputs = self._outputs().get()
        outputs["subject"] = self.inputs.subject
        outputs["study_dir"] = self.inputs.study_dir
        return outputs

class OutputSelectorInputSpec(BaseInterfaceInputSpec):
    study_dir = Directory(
            desc="The study dir. Should contain a subdir named after the subject.")
    output_mni_only = traits.Bool(True,
            mandatory=False, usedefault=True,
            desc = """
            To save space, by default we only output the pipeline's final
            product. If you need to look at the precursor files, then just set
            this to false... and prepare your drives to go to that big server
            farm in the sky.
            """)

class OutputSelectorOutputSpec(TraitedSpec):
    output_dir = Directory(
            desc="Either the full study dir, or just the MNINonLinear dir (default).")

class OutputSelector(BaseInterface):
    input_spec = OutputSelectorInputSpec
    output_spec = OutputSelectorOutputSpec
    out_dir = None
    def _run_interface(self, runtime):
        from glob import glob
        od = self.inputs.study_dir
        if not self.inputs.output_mni_only:
            self.out_dir = od
        else:
            glob_str = os.path.join(od, '*', 'MNINonLinear')
            mni_dirs = glob(glob_str)
            if not mni_dirs:
                raise ValueError("Could not find MNINonLinear output dir in %s" % glob_str)
            self.out_dir = mni_dirs[0]
        return runtime

    def _list_outputs(self):
        from glob import glob
        outputs = self._outputs().get()
        outputs["output_dir"] = self.out_dir
        return outputs







import os
import sys
from util import *
from multiprocessing import Pool, cpu_count
from configobj import ConfigObj

SCAN_TYPES = [
    "bold",
    "bold_sbref",
    "fieldmap_magnitude",
    "fieldmap_phase",
    "fieldmap_ap",
    "fieldmap_lr",
    "fieldmap_pa",
    "fieldmap_rl",
    "t1",
    "t2",
    "polarity_swapped"]

TEMPL_KEYS = [
    "templates_dir",
    "t1_template",
    "t1_template_brain",
    "t1_template_2mm",
    "t2_template",
    "t2_template_brain",
    "t2_template_2mm",
    "template_mask",
    "template_2mm_mask",
    "surf_atlas_dir",
    "grayordinates_dir",
    "ref_myelin_maps",
    ]

CONF_FILE_KEYS = [
    "fnirt_config",
    "grad_distort_coeffs",
    "subcort_gray_labels",
    "freesurfer_labels",
    "top_up_config",]

CONF_FILE_DEFAULTS = [
    "%(hcp_dir)s/global/config/T1_2_MNI152_2mm.cnf",
    "%(hcp_dir)s/global/config/coeff_SC72C_Skyra.grad",
    "%(hcp_dir)s/global/config/FreeSurferSubcorticalLabelTableLut.txt",
    "%(hcp_dir)s/global/config/FreeSurferAllLut.txt",
    "%(hcp_dir)s/global/config/b02b0.cnf",]

POS_FIELDMAPS = ["fieldmap_rl", "fieldmap_pa"]
NEG_FIELDMAPS = ["fieldmap_lr", "fieldmap_ap"]
YES_WORDS = [1,"1","y","Y","yes","Yes","YES", "True", "true"]
NO_WORDS = [0,"0","n","N","no","No","NO", "False", "false"]
EPFM_SELECTION_OPTS = ["first","most_recent"]

def setup_conf():
    import os
    # get name for conf file
    name = raw_input("New config file name [hcp.conf]: ")
    name = name.strip()
    name = name if name else "hcp.conf"
    name = name if (len(name) > 5 and name[-5:] == ".conf") else name + ".conf"
    # make sure that file doesn't already exist
    if os.path.exists(name):
        print "File already exists."
        return
    # get the directory containing all data for all subjects
    print "\nThe subjects directory should contain all raw data for all subjects."
    subs_dir = raw_input("Subjects Directory [./]: ")
    subs_dir = subs_dir.strip()
    subs_dir = subs_dir if subs_dir else "./"
    subs_dir = os.path.abspath(subs_dir)
    # get the template used for getting form the subject dir to the dicoms
    print "\nThe DICOM template should be a format string for a glob which, " + \
          "when combined with an individual subject ID, will get us all of " + \
          "the subject's DICOM files."
    dcm_temp = raw_input("DICOM template [data/raw_dicom/%s/*.dcm]: ")
    dcm_temp = dcm_temp.strip()
    dcm_temp = dcm_temp if dcm_temp else "data/raw_dicom/%s/*.dcm"
    dcm_temp = dcm_temp if ".dcm" in dcm_temp else os.path.join(dcm_temp, "*.dcm")
    # get a list of subjects
    print "\nSubjects should be a comma separated list of subject ids."
    subs = raw_input("Subject list ['']: ")
    subs = subs.strip()
    # set up config obj
    config = ConfigObj(name, unrepr=True)
    # get the basics
    config["general"] = {}
    config["general"]["subjects"] = [s.strip() for s in subs.split(",")]
    config["general"]["subject_dir"] = subs_dir
    config["general"]["dicom_template"] = dcm_temp
    # write the config file
    config.write()
    update_conf(name)

def get_series_desc(dicom_path):
    import dicom
    d = dicom.read_file(dicom_path)
    sd = getattr(d, "SeriesDescription", None)
    return sd

def templ_defaults_for_rez(rez):
    if rez in [.7, .8]:
        rez = "%0.1f" % rez
    elif rez in [1]:
        rez = "%1.f" % rez
    else:
        return None
    return [
        "%(hcp_dir)s/global/templates",
        "%%(templates_dir)s/MNI152_T1_%smm.nii.gz" % rez,
        "%%(templates_dir)s/MNI152_T1_%smm_brain.nii.gz" % rez,
        "%(templates_dir)s/MNI152_T1_2mm.nii.gz",
        "%%(templates_dir)s/MNI152_T2_%smm.nii.gz" % rez,
        "%%(templates_dir)s/MNI152_T2_%smm_brain.nii.gz" % rez,
        "%(templates_dir)s/MNI152_T2_2mm.nii.gz",
        "%%(templates_dir)s/MNI152_T1_%smm_brain_mask.nii.gz" % rez,
        "%(templates_dir)s/MNI152_T1_2mm_brain_mask_dil.nii.gz",
        "%(templates_dir)s/standard_mesh_atlases",
        "%(templates_dir)s/91282_Greyordinates",
        "%(templates_dir)s/standard_mesh_atlases/Conte69.MyelinMap_BC.164k_fs_LR.dscalar.nii"]

def update_conf(conf_path):
    import os
    import sys
    from glob import glob
    import dicom
    from time import sleep
    from numpy import unique
    # TODO: make sure conf is there
    config = ConfigObj(conf_path, unrepr=True)
    # update the series map if needed or desired
    needs_smap = "series" not in config
    if not needs_smap:
        needs_smap = raw_input("Do you want to re-map the series descriptions y/[n]? ").strip() in YES_WORDS
    if needs_smap:
        # get list of all sequence descriptions
        s = config["general"]["subject_dir"]
        t = config["general"]["dicom_template"] % "*"
        dicoms = glob(os.path.join(s,t))
        message = "\rChecking series names (this may take some time) %s"
        dcm_count = len(dicoms)
        pool = Pool(processes=min(15, int(round(cpu_count() * .75))))
        result = pool.map_async(get_series_desc, dicoms)
        while not result.ready():
            sleep(.5)
            # perc = float(dcm_count - result._number_left) / float(dcm_count)
            # perc = "(%d%%)..." % int(round(perc * 100))
            perc = "(%d chunks remaining)..." % result._number_left
            m = message % perc
            sys.stdout.write(m)
            sys.stdout.flush()
        series = unique(result.get()).tolist()
        if None in series:
            series.remove(None)
        found = "\nFound %d unique series descriptions.\n" % len(series)
        sys.stdout.write(found)
        # message if we didn't find anything
        if not series:
            raise ValueError("couldn't find any dicoms! please double check your paths and templates...")
        # update the series confg
        series.sort()
        print "-------\nSeries:\n-------"
        print "\n".join(["%d:\t%s" % (i, ser) for i,ser in enumerate(series)]) + "\n"
        type_matches = dict(zip(SCAN_TYPES, [[] for s in SCAN_TYPES]))
        i = 0
        series_count = len(series)
        while i < len(SCAN_TYPES):
            ft = SCAN_TYPES[i]
            m = raw_input("\nWhich series do you use for '%s'?\n[None] or comma separated values 0-%d: " % (ft, len(series)-1)).strip()
            if not m:
                i += 1
                continue
            try:
                m = [int(n) for n in m.split(",")]
                if any([n < 0 or n >= series_count for n in m]):
                    raise ValueError()
            except Exception, e:
                print "Invalid Selection..."
                continue
            i += 1
            for n in m:
                type_matches[ft].append(series[n])
        config["series"] = {}
        for key in sorted(type_matches.keys()):
            config["series"][key] = type_matches[key]
    config["DEFAULT"] = {}
    # freesurfer home (defaults)
    d_val = os.environ.get("FREESURFER_HOME", "")
    fsh = raw_input("\nPath for FREESURFER_HOME [%s]: " % d_val).strip()
    fsh = fsh if fsh else d_val
    config["DEFAULT"]["freesurfer_home"] = fsh
    # fsl dir (defaults)
    d_val = os.environ.get("FSLDIR", "")
    fslh = raw_input("\nPath for FSLDIR [%s]: " % d_val).strip()
    fslh = fslh if fslh else d_val
    config["DEFAULT"]["fsl_dir"] = fslh
    # hcp dir (defaults)
    hcph = raw_input("\nPath to HCP Pipelines dir [ ]: ").strip()
    config["DEFAULT"]["hcp_dir"] = hcph
    # template file locations (templates)
    rez = raw_input("\nWhat is your structural image resolution (mm)?\n[Skip] or one of (.7, .8, 1): ").strip()
    rez = float_or_none(rez)
    rez_temps = None
    def_t = False
    if rez:
        rez_temps = templ_defaults_for_rez(rez)
    if rez_temps:
        def_t = raw_input("\nUse default template files for resolution %0.1f [y]/n? " % rez).strip() not in NO_WORDS
        if def_t:
            t_vals = rez_temps
    if not t_vals:
        t_vals = ['' for k in TEMPL_KEYS]
    config["templates"] = {}
    for i, key in enumerate(TEMPL_KEYS):
        config["templates"][key] = t_vals[i]
    if not def_t:
        print "\nWhen finished, please open your config file to set the tamplate locations manually.\n"
    # config file locations (config_files)
    def_c = raw_input("\nUse default config files [y]/n?").strip() not in NO_WORDS
    c_vals = CONF_FILE_DEFAULTS if def_c else ['' for k in CONF_FILE_KEYS]
    config["config_files"] = {}
    for i, key in enumerate(CONF_FILE_KEYS):
        config["config_files"][key] = c_vals[i]
    if not def_c:
        print "\nWhen finished, please open your config file to set the other config file locations manually.\n"
    config["nifti_wrangler"] = {}
    # ep unwarp policy
    print "\nIf you have more than one ep fieldmap set, you may either\nwant to use the first, or always use the most recent."
    ep_sel = numbered_choice(EPFM_SELECTION_OPTS, prompt="Which policy would you like to use:")
    config["nifti_wrangler"]["ep_fieldmap_selection"] = ep_sel
    # struct averaging block
    print "\nIf you collect multiple t1 or t2 images, and averaging them yields warped\nresults, try blocking structural image averaging."
    block_avg = raw_input("\nBlock averaging of structural images [y]/n? ").strip() not in NO_WORDS
    config["nifti_wrangler"]["block_struct_averaging"] = block_avg
    # a terrible hack, but take a guess at the unwarp direction
    uwd = "x"
    ser = config["series"]
    if ser.get("fieldmap_ap",None) and ser.get("fieldmap_pa",None):
        uwd = "y"
    elif ser.get("fieldmap_rl",None) and ser.get("fieldmap_lr",None):
        uwd = "x"
    config["nifti_wrangler"]["ep_unwarp_dir"] = uwd
    print "\nVery weak guess that your primary unwarp direction is %s.\nDid I mention this is a GUESS?" % uwd
    print "\nWhen finished, please open your config file check the value for ep_unwarp_dir.\n"
    # output level selection
    print """
    By default, we only output the MNINonLinear results folder (recommended, for
    the sake of your drive capacity). If you want to check out the full outputs,
    including all of the precursor files, set output_mni_only to False in the
    config file."""
    config["output_select"] = {"output_mni_only":True}
    # write the file!
    config.write()

def select_conf():
    # if there"s only one conf around, select it. otherwise, offer a choice.
    from glob import glob
    confs = glob("./*.conf")
    if not confs:
        raise ValueError("Could not find any .conf files in current directory.")
    if len(confs) == 1:
        return confs[0]
    return numbered_choice(confs, prompt = "There were multiple config files. Please choose.")

def validate_config(conf_dict):
    """ validates the config dict
    return: True if it passes, False if it fails
    """
    # TODO: more of this
    # make sure we can get an env out of it
    try:
        get_hcp_env_for_config(conf_dict)
        get_hcp_commands_for_config(conf_dict)
    except Exception, e:
        return False
    return True

def numbered_choice(choices, prompt=None, allow_multiple=False, allow_none=False):
    import numpy as np
    # TODO: print a choice, return the chosen value
    if not prompt:
        prompt = "Options:\n" + "-"*len("Options:")
    print prompt + "\n"
    for i, val in enumerate(choices):
        print "%d: %s" % (i, val)
    # print "\n"
    nstr = "[None] or " if allow_none else ""
    mstr = "comma separated values " if allow_multiple else ""
    selection = None
    done = False
    while not done:
        try:
            m = raw_input("\nSelect %s%s0-%d: " % (nstr, mstr, len(choices)-1)).strip()
            if not m:
                if allow_none:
                    selection = None
                    done = True
                else:
                    raise ValueError()
            else:
                m = [int(n) for n in m.split(",")]
                if len(m) > 1 and not allow_multiple:
                    raise ValueError()
                if any([n < 0 or n >= len(choices) for n in m]):
                    raise ValueError()
                m = np.unique(m)
                selection = [choices[n] for n in m]
                if not allow_multiple:
                    selection = selection[0]
                done = True
        except ValueError, e:
            print "Invalid Selection..."
            continue
    return selection

def get_config_dict(conf_path):
    config = ConfigObj(conf_path, unrepr=True)
    return config

def get_hcp_env_for_config(conf_dict):
    # exceptions will be raised if config isn't good.
    # hint: we call this from the validation method!
    d = conf_dict["DEFAULT"]
    dp = lambda x: os.path.join(d["hcp_dir"], x)
    return {
        "FSLDIR":d["fsl_dir"],
        "FREESURFER":d["freesurfer_home"],
        "HCPPIPEDIR":d["hcp_dir"],
        "CARET7DIR":dp("global/binaries/caret7/bin_rh_linux64"),
        "HCPPIPEDIR_Templates":dp("global/templates"),
        "HCPPIPEDIR_Bin":dp("global/binaries"),
        "HCPPIPEDIR_Config":dp("global/config"),
        "HCPPIPEDIR_PreFS":dp("PreFreeSurfer/scripts"),
        "HCPPIPEDIR_FS":dp("FreeSurfer/scripts"),
        "HCPPIPEDIR_PostFS":dp("PostFreeSurfer/scripts"),
        "HCPPIPEDIR_fMRISurf":dp("fMRISurface/scripts"),
        "HCPPIPEDIR_fMRIVol":dp("fMRIVolume/scripts"),
        "HCPPIPEDIR_tfMRI":dp("tfMRI/scripts"),
        "HCPPIPEDIR_dMRI":dp("DiffusionPreprocessing/scripts"),
        "HCPPIPEDIR_dMRITract":dp("DiffusionTractography/scripts"),
        "HCPPIPEDIR_Global":dp("global/scripts"),
        "HCPPIPEDIR_tfMRIAnalysis":dp("TaskfMRIAnalysis/scripts"),
        "MSMBin":dp("MSMBinaries"),
        }

def get_hcp_commands_for_config(conf_dict):
    # exceptions will be raised if config isn't good.
    # hint: we call this from the validation method!
    dp = lambda x: os.path.join(conf_dict["DEFAULT"]["hcp_dir"], x)
    return [
        dp("PreFreeSurfer/PreFreeSurferPipeline.sh"),
        dp("FreeSurfer/FreeSurferPipeline.sh"),
        dp("PostFreeSurfer/PostFreeSurferPipeline.sh"),
        dp("fMRIVolume/GenericfMRIVolumeProcessingPipeline.sh"),
        dp("fMRISurface/GenericfMRISurfaceProcessingPipeline.sh"),
        ]

def apply_dict_to_obj(the_d, obj, skip_names=[]):
    if not the_d:
        return
    for name, val in the_d.iteritems():
        if name in skip_names or "traits" not in dir(obj) or name not in obj.traits().keys():
            continue
        setattr(obj, name, val)



#!/usr/bin/env python
# encoding: utf-8

"""
hcpre.py

Copyright (c) 2013 Ben Acland
"""

import os
import sys
import getopt

from workflows import *

help_message = """
Provides command line tools to build and work on workflow confiuration files,
and for launching the pipeline using a specific config. You can either import
HCPrepWorkflow, then configure and run in from your own code, or use the
built-in command line tools to launch the workflow on your data.

Commands
--------
-h, --help
    Prints out this message.
-i, --init
    Creates a new config file in your current directory.
-u, --update
    Reruns part of the config file setup script on any config file in your current directory.
-g, --graph
    Draws a graph of the nipype workflow to the current directory.
-r, --run
    Runs the workflow using any config file in the current directory.

Parameters
----------
-c, --config (path)
    The config file to use.
-s (comma separated subject numbers)
    The list of subjects who you'd like to put through the workflow. Overrides
    setting in the config file.
-n (integer)
    The number of threads you would like to use. Higher numbers are faster to
    a point, but depending on how large your data is (chances are, it is quite
    large), you may well want to limit yourself to something below 8 for
    starters if you're working on a large server. Default is 1. Ignored if
    you use -p.
-p, --pbs
    Causes nipype to try to use the PBS plugin. For use on the cluster only.
    Experimental.
-o (path)
    The directory to put preprocessed data in. Default is current directory.
-v 
    Verbose. At this point, just makes -g print a graph that expands iterables.
"""

class Usage(Exception):
    def __init__(self, msg=help_message):
        self.msg = msg

def main(argv=None):
    import os
    if argv is None:
        argv = sys.argv
    try:
        try:
            opts, args = getopt.getopt(argv[1:], "pvrhiugs:n:o:c:", ["run", "help", "init", "update", "graph", "config", "pbs"])
        except getopt.error, msg:
            raise Usage(msg="\n"+str(msg))
        # option processing
        update = False
        graph = False
        subs = None
        N_PROCS = None
        run = False
        verbose = False
        c_file = None
        use_pbs = False
        out_dir = os.getcwd()
        for option, value in opts:
            if option in ("-h", "--help"):
                raise Usage()
            if option in ("-i", "--init"):
                setup_conf()
                return
            if option in ("-u", "--update"):
                update = True
            if option in ("-g", "--graph"):
                graph = True
            if option in ("-r", "--run"):
                run = True
            if option in ("-c", "--config"):
                c_file = value
            if option in ("-p", "--pbs"):
                use_pbs = True
            if option in ("-s"):
                subs = [sub.strip() for sub in value.split(",")]
            if option in ("-v"):
                verbose = True
            if option in ('-n'):
                N_PROCS = int_or_none(value)
            if option in ('-o'):
                out_dir = value
        # select config file
        if not c_file:
            try:
                c_file = select_conf()
            except Exception, e:
                raise Usage(msg="Could not find a config file.")
        # update if necessary
        if update:
            update_conf(c_file)
            return
        # make sure we're going to do something
        if not run and not graph:
            raise Usage(msg="Nothing to do...")
        # validate the config for running
        conf = get_config_dict(c_file)
        if not conf:
            raise Usage(msg="Could not parse config file.")
        if not validate_config(conf):
            raise Usage(msg="Invalid config file.")
        # build the workflow, pass it subjects if they were given in the command line
        wk = HCPrepWorkflow(name="hcp_prep_workflow", config=conf)
        if subs:
            wk.subjects_node.iterables = ("subject", subs)
        # set the output dir
        out_dir = os.path.abspath(out_dir)
        if not os.path.exists(out_dir):
            os.makedirs(out_dir)
        wk.data_sink.inputs.base_directory = out_dir
        # graph if you like
        if graph:
            g2u = "exec" if verbose else "orig"
            wk.write_graph(dotfilename="hcp_pipe_graph", graph2use=g2u)
            return
        if not run:
            return
        if use_pbs:
            print "running using PBS"
            # note - the default qsub args are very modest. use this structure
            # to scale up if necessary.
            ques = [
                [[wk.dicom_convert, wk.dicom_info, wk.nii_wrangler],
                    {"qsub_args":"-l nodes=1:ppn=1,mem=1gb,walltime=1:00:00"}],
                [[wk.hc_pre_fs],
                    {"qsub_args":"-l nodes=1:ppn=2,mem=10gb,vmem=10gb,walltime=6:00:00"}],
                [[wk.hc_fs],
                    {"qsub_args":"-l nodes=1:ppn=2,mem=5gb,walltime=24:00:00"}],
                [[wk.hc_post_fs],
                    {"qsub_args":"-l nodes=1:ppn=4,mem=10gb,walltime=4:00:00"}],
                [[wk.hc_volume, wk.hc_surface],
                    {"qsub_args":"-l nodes=1:ppn=4,mem=10gb,walltime=12:00:00"}],
                ]
            for q in ques:
                p_args = dict(q[1],**{"overwrite":True})
                for n in q[0]:
                    n.plugin_args = p_args
            wk.run(plugin="PBS",
                   plugin_args={"qsub_args":"-l nodes=1:ppn=1,mem=1gb,walltime=1:00:00"})
        elif N_PROCS > 0:
            print "running with %d processes" % N_PROCS
            wk.run(plugin="MultiProc", plugin_args={"n_procs" : N_PROCS, "non_daemon" : True})
        else:
            print "running single process"
            wk.run()
    except Usage, err:
        f_str = sys.argv[0].split("/")[-1] + ":"
        lfs = len(f_str)
        f_str = "%s\n%s\n%s\n" % ("-"*lfs, f_str, "-"*lfs)
        print >> sys.stderr, f_str + str(err.msg)
        print >> sys.stderr, "-------------------\nfor help use --help\n-------------------"
        return 2

if __name__ == "__main__":
    sys.exit(main())




import nipype.interfaces.io as nio           # Data i/o
import nipype.interfaces.utility as util     # utility
import nipype.pipeline.engine as pe          # pypeline engine

from interfaces import *
from config import *
from util import *

# TODO: validate these guesses... user may need to look at an image and flip an option if necessary
# Positive fieldmaps: RL, AP
# Negative fieldmaps: LR, PA

# TODO: update to account for the scalar -> list changes in HCP code (prefs, I think...)

# TODO: use the following values to select the correct template resolution?
# (0028, 0030) Pixel Spacing                       DS: ['0.80000001192093', '0.80000001192093']
# (0018, 0050) Slice Thickness                     DS: '0.79999995231628'

class HCPrepWorkflow(pe.Workflow):
    def __init__(self, config=None, *args, **kwargs):
        super(HCPrepWorkflow, self).__init__(*args, **kwargs)
        self.hc_config = config

    @property
    def hc_config(self):
        return self._hc_config
    @hc_config.setter
    def hc_config(self, value):
        self._hc_config = value
        if self._hc_config:
            self.update_nodes_from_config()

    def get_conf(self, section, option):
        if not self.hc_config:
            return None
        return self.hc_config.get(section,{}).get(option, None)

    def update_nodes_from_config(self):
        # subjects node
        subs = self.get_conf("general","subjects")
        if subs:
            self.subjects_node.iterables = ("subject", subs)
        # dcm grabber
        sub_dir = self.get_conf("general","subject_dir")
        dcm_temp = self.get_conf("general","dicom_template")
        if sub_dir:
            self.dicom_grabber.inputs.base_directory = sub_dir
        if dcm_temp:
            self.dicom_grabber.inputs.field_template = {"dicom": dcm_temp}
        # nifti wrangler
        series_map = self.hc_config.get("series", {})
        if series_map:
            self.nii_wrangler.inputs.series_map = series_map
        # set template and config values (names are also input names on some nodes)
        temps = self.hc_config.get("templates", {})
        c_files = self.hc_config.get("config_files", {})
        for n in [self.hc_pre_fs, self.hc_fs, self.hc_post_fs, self.hc_volume, self.hc_surface]:
            apply_dict_to_obj(temps, n.inputs, skip_names=["templates_dir"])
            apply_dict_to_obj(c_files, n.inputs)
        # set the commands for each of the freesurfer steps... this stinks. improve when you can.
        (self.hc_pre_fs.inputs.full_command,
            self.hc_fs.inputs.full_command,
            self.hc_post_fs.inputs.full_command,
            self.hc_volume.inputs.full_command,
            self.hc_surface.inputs.full_command) = get_hcp_commands_for_config(self.hc_config)
        # set the environ on each of the hc nodes
        envs = get_hcp_env_for_config(self.hc_config)
        self.hc_pre_fs.inputs.environ = envs
        self.hc_fs.inputs.environ = envs
        self.hc_post_fs.inputs.environ = envs
        self.hc_volume.inputs.environ = envs
        self.hc_surface.inputs.environ = envs
        # any other per-step hcp config - a good place to overide un-derived values
        apply_dict_to_obj(self.hc_config.get("nifti_wrangler", {}), self.nii_wrangler.inputs)
        apply_dict_to_obj(self.hc_config.get("pre_freesurfer", {}), self.hc_pre_fs.inputs)
        apply_dict_to_obj(self.hc_config.get("freesurfer", {}), self.hc_fs.inputs)
        apply_dict_to_obj(self.hc_config.get("post_freesurfer", {}), self.hc_post_fs.inputs)
        apply_dict_to_obj(self.hc_config.get("volume_processing", {}), self.hc_volume.inputs)
        apply_dict_to_obj(self.hc_config.get("surface_processing", {}), self.hc_surface.inputs)
        apply_dict_to_obj(self.hc_config.get("output_select", {}), self.output_select.inputs)

    def run(self, *args, **kwargs):
        self.connect_nodes()
        super(HCPrepWorkflow, self).run(*args, **kwargs)

    def write_graph(self, *args, **kwargs):
        self.connect_nodes()
        super(HCPrepWorkflow, self).write_graph(*args, **kwargs)
        
    def clear_nodes(self):
        all_nodes = self._get_all_nodes()
        if all_nodes is not None:
            self.remove_nodes(all_nodes)

    def connect_nodes(self):
        # Some connections that don't change
        self.clear_nodes()
        self.connect([
            # prep steps
            (self.subjects_node, self.dicom_grabber, [("subject", "subject")]),
            (self.dicom_grabber, self.dicom_gather, [("dicom", "files")]),
            (self.dicom_gather, self.dicom_select, [("links", "inlist")]),
            (self.dicom_select, self.dicom_convert, [("out", "source_names")]),
            (self.dicom_gather, self.dicom_info, [("links", "files")]),
            (self.dicom_convert, self.nii_wrangler, [("converted_files", "nii_files")]),
            (self.dicom_info, self.nii_wrangler, [("info", "dicom_info")]),
            # pre freesurfer
            (self.subjects_node, self.hc_pre_fs, [("subject", "subject")]),
            (self.nii_wrangler, self.hc_pre_fs, [("t1_structs", "t1_files")]),
            (self.nii_wrangler, self.hc_pre_fs, [("t2_structs", "t2_files")]),
            (self.nii_wrangler, self.hc_pre_fs, [("mag_fieldmap", "fieldmap_magnitude")]),
            (self.nii_wrangler, self.hc_pre_fs, [("phase_fieldmap", "fieldmap_phase")]),
            (self.nii_wrangler, self.hc_pre_fs, [("fieldmap_te", "fieldmap_te")]),
            (self.nii_wrangler, self.hc_pre_fs, [("t1_sample_spacing", "t1_sample_spacing")]),
            (self.nii_wrangler, self.hc_pre_fs, [("t2_sample_spacing", "t2_sample_spacing")]),
            # freesurfer
            (self.hc_pre_fs, self.hc_fs, [("subject", "subject")]),
            (self.hc_pre_fs, self.hc_fs, [("subject_t1_dir", "subject_t1_dir")]),
            (self.hc_pre_fs, self.hc_fs, [("t1_acpc_dc_restore", "t1_acpc_dc_restore")]),
            (self.hc_pre_fs, self.hc_fs, [("t1_acpc_dc_restore_brain", "t1_acpc_dc_restore_brain")]),
            (self.hc_pre_fs, self.hc_fs, [("t2_acpc_dc_restore", "t2_acpc_dc_restore")]),
            # post freesurfer
            (self.hc_fs, self.hc_post_fs, [("subject", "subject")]),
            (self.hc_pre_fs, self.hc_post_fs, [("study_dir", "study_dir")]),
            # volume
            (self.hc_post_fs, self.hc_volume, [("subject", "subject")]),
            (self.hc_pre_fs, self.hc_volume, [("study_dir", "study_dir")]),
            (self.nii_wrangler, self.hc_volume, [("bold_names", "bold_name")]),
            (self.nii_wrangler, self.hc_volume, [("bolds", "bold_img")]),
            (self.nii_wrangler, self.hc_volume, [("sb_refs", "bold_scout")]),
            (self.nii_wrangler, self.hc_volume, [("neg_fieldmaps", "se_fieldmap_neg")]),
            (self.nii_wrangler, self.hc_volume, [("pos_fieldmaps", "se_fieldmap_pos")]),
            (self.nii_wrangler, self.hc_volume, [("ep_echo_spacings", "fieldmap_echo_spacing")]),
            (self.nii_wrangler, self.hc_volume, [("ep_unwarp_dirs", "unwarp_dir")]),
            (self.hc_post_fs, self.hc_volume, [("grayordinates_res", "fmri_res")]),
            # surface
            (self.hc_volume, self.hc_surface, [("subject", "subject")]),
            (self.hc_volume, self.hc_surface, [("bold_name", "bold_name")]),
            (self.hc_pre_fs, self.hc_surface, [("study_dir", "study_dir")]),
            (self.hc_post_fs, self.hc_surface, [("low_res_mesh", "low_res_mesh")]),
            (self.hc_post_fs, self.hc_surface, [("grayordinates_res", "fmri_res")]),
            (self.hc_post_fs, self.hc_surface, [("grayordinates_res", "grayordinates_res")]),
            # data join and sink
            (self.hc_surface, self.data_join, [("study_dir", "inlist")]),
            (self.data_join, self.output_select, [("out", "study_dir")]),
            (self.output_select, self.data_sink, [("output_dir", "preprocessed")]),
            ])

    """ self-inflating nodes """
    
    @property
    def subjects_node(self):
        if not getattr(self,"_subjects_node",None):
            self._subjects_node = pe.Node(
                    name="subs_node",
                    interface=util.IdentityInterface(
                            fields=["subject"]))
        return self._subjects_node
    @subjects_node.setter
    def subjects_node(self, val):
        self._subjects_node = val

    @property
    def dicom_grabber(self):
        if not getattr(self,"_dicom_grabber",None):
            self._dicom_grabber = pe.Node(
                    name = "dicom_source_1",
                    interface = nio.DataGrabber(
                            infields = ["subject"],
                            outfields = ["dicom"],))
            self._dicom_grabber.inputs.template = "*"
            self._dicom_grabber.inputs.template_args = {"dicom": [["subject"]]}
            self._dicom_grabber.inputs.sort_filelist = True
        return self._dicom_grabber
    @dicom_grabber.setter
    def dicom_grabber(self, val):
        self._dicom_grabber = val

    @property
    def dicom_gather(self):
        if not getattr(self,'_dicom_gather',None):
            self._dicom_gather = pe.Node(name='dicom_gather', interface=GatherFiles())
        return self._dicom_gather
    @dicom_gather.setter
    def dicom_gather(self, val):
        self._dicom_gather = val

    @property
    def dicom_convert(self):
        if not getattr(self,"_dicom_convert",None):
            self._dicom_convert = pe.Node(name="dicom_convert", interface=HCDcm2nii())
            self._dicom_convert.inputs.convert_all_pars = True
            self._dicom_convert.inputs.gzip_output = False
            self._dicom_convert.inputs.reorient = False
            self._dicom_convert.inputs.reorient_and_crop = False
            self._dicom_convert.inputs.args = "-d n -p n"
        return self._dicom_convert
    @dicom_convert.setter
    def dicom_convert(self, val):
        self._dicom_convert = val

    @property
    def dicom_select(self):
        if not getattr(self,'_dicom_select',None):
            self._dicom_select = pe.Node(name="select_dicom", interface=util.Select(index = 0))
        return self._dicom_select
    @dicom_select.setter
    def dicom_select(self, val):
        self._dicom_select = val

    @property
    def dicom_info(self):
        if not getattr(self,'_dicom_info',None):
            self._dicom_info = pe.Node(name="dicom_info", interface=DicomInfo())
        return self._dicom_info
    @dicom_info.setter
    def dicom_info(self, val):
        self._dicom_info = val

    @property
    def nii_wrangler(self):
        if not getattr(self,'_nii_wrangler',None):
            self._nii_wrangler = pe.Node(name="nii_wrangler", interface=NiiWrangler())
        return self._nii_wrangler
    @nii_wrangler.setter
    def nii_wrangler(self, val):
        self._nii_wrangler = val

    @property
    def hc_pre_fs(self):
        if not getattr(self,'_hc_pre_fs',None):
            self._hc_pre_fs = pe.Node(name="pre_freesurfer", interface=PreFS())
        return self._hc_pre_fs
    @hc_pre_fs.setter
    def hc_pre_fs(self, val):
        self._hc_pre_fs = val

    @property
    def hc_fs(self):
        if not getattr(self,'_hc_fs',None):
            self._hc_fs = pe.Node(name="freesurfer", interface=FS())
        return self._hc_fs
    @hc_fs.setter
    def hc_fs(self, val):
        self._hc_fs = val

    @property
    def hc_post_fs(self):
        if not getattr(self,'_hc_post_fs',None):
            self._hc_post_fs = pe.Node(name="post_freesurfer", interface=PostFS())
        return self._hc_post_fs
    @hc_post_fs.setter
    def hc_post_fs(self, val):
        self._hc_post_fs = val

    @property
    def hc_volume(self):
        if not getattr(self,'_hc_volume',None):
            self._hc_volume = pe.MapNode(
                    name="volume",
                    interface=VolumeProcessing(),
                    iterfield=["bold_name",
                               "bold_img",
                               "bold_scout",
                               "se_fieldmap_pos",
                               "se_fieldmap_neg",
                               "unwarp_dir",
                               "fieldmap_echo_spacing"])
        return self._hc_volume
    @hc_volume.setter
    def hc_volume(self, val):
        self._hc_volume = val

    @property
    def hc_surface(self):
        if not getattr(self,'_hc_surface',None):
            self._hc_surface = pe.MapNode(
                    name="surface",
                    interface=SurfaceProcessing(),
                    iterfield=["bold_name",
                               "subject"]) # subject, because this gets passed from another mapnode...
        return self._hc_surface
    @hc_surface.setter
    def hc_surface(self, val):
        self._hc_surface = val

    @property
    def data_join(self):
        if not getattr(self,'_data_join',None):
            self._data_join = pe.Node(name="data_join", interface=util.Select(index = 0))
        return self._data_join
    @data_join.setter
    def data_join(self, val):
        self._data_join = val

    @property
    def output_select(self):
        if not getattr(self,'_output_select',None):
            self._output_select = pe.Node(name="output_select", interface=OutputSelector())
        return self._output_select
    @output_select.setter
    def output_select(self, val):
        self._output_select = val

    @property
    def data_sink(self):
        if not getattr(self,'_data_sink',None):
            self._data_sink = pe.Node(name="data_sink", interface=nio.DataSink())
        return self._data_sink
    @data_sink.setter
    def data_sink(self, val):
        self._data_sink = val


def orientation_from_dcm_header(header):
    # borrowed from hcp xnat dicom juggling java
    # get and check the base values
    if not header:
        raise ValueError("didn't get a header")
    o = getattr(header, "ImageOrientationPatient", None)
    o = [float(a) for a in o]
    if not o:
        raise ValueError("couldn't find ImageOrientationPatient in header")
    if len(o) != 6:
        raise ValueError("cannot be translated to cosine vectors")
    # consistency checks
    epsilon = 0.001
    if abs(o[0] * o[3] + o[1] * o[4] + o[2] * o[5]) > 0.001:
        raise ValueError("cosine vectors not orthogonal")
    if abs(1.0 - o[0] * o[0] - o[1] * o[1] - o[2] * o[2]) > epsilon:
        raise ValueError("cosine vectors not normal")
    # looks like we're good to go. derive the value
    absNormalX = abs(o[1] * o[5] - o[2] * o[4])
    absNormalY = abs(o[2] * o[3] - o[0] * o[5])
    absNormalZ = abs(o[0] * o[4] - o[1] * o[3])
    if absNormalX > absNormalY:
        return "sagittal" if absNormalX > absNormalZ else "transverse"
    else:
        return "coronal" if absNormalY > absNormalZ else "transverse"

def numberfy(s):
    n = s
    try:
        n = float(n)
        return n
    except Exception:
        return s

def float_or_none(s):
    n = s
    try:
        n = float(n)
        return n
    except Exception:
        return None

def int_or_none(s):
    n = s
    try:
        n = int(n)
        return n
    except ValueError:
        return None


"""A home for constants."""
# Python modules
from __future__ import division

# 3rd party modules

# Our modules
# import vespa.common.ordered_dict as ordered_dict
from collections import OrderedDict as ordered_dict

# We could also use numpy.pi in place of 3.14..., but it's useful to limit 
# this module's dependencies.
DEGREES_TO_RADIANS = 3.1415926535897931 / 180
RADIANS_TO_DEGREES = 180 / 3.1415926535897931


class PaneNames(object):
    """ A container that allows me to represent wxWidgets pane names (which
    are strings) in a way that Python will complain loudly about if I mistype
    one of them (e.g. psuedo instead of pseudo). A pane's name is its unique
    handle. They are not exposed to the user and they are not magic. (i.e.
    the name "baseline" could just as well be "shrubbery".)
    """
    BASELINE = "baseline"
    XYZ_BROWSER = "xyz_browser"
    SUM_SPECTRA = "sum_spectra"
    TIME_SERIES = "time_series"
    KORNAK = "kornak"
    LORENTZ_GAUSS = "lorentz_gauss"
    PSEUDO_2D = "pseudo_2d"
    HEADER = "header"
    SPECTRAL = "spectral"
    SPATIAL = "spatial"
    FILEBAR = "filebar"
    SVD = "svd"
    FITTING = "fitting"
    PROCESSING = "processing"
    
    TOOLS = ( FILEBAR, SPATIAL, BASELINE, XYZ_BROWSER, SUM_SPECTRA, 
              TIME_SERIES, KORNAK, LORENTZ_GAUSS, PSEUDO_2D)


class AmplitudeMultiplier(object):
    """ Amplitude multiplier constants """
    MIN = 0
    MAX = 1e12
    
class Apodization(object):
    """ Apodization constants """
    MIN_WIDTH =   0
    MAX_WIDTH = 100

    # These constants are arbitrary and may change.
    # However bool(NONE) is guaranteed to be False
    NONE = 0
    GAUSSIAN = 1
    LORENTZIAN = 2

    # Items for the spectral processing options dropdown
    choices = ordered_dict(   (   (NONE , "None"), 
                                        (GAUSSIAN , "Gaussian"),
                                        (LORENTZIAN , "Lorentzian"),
                                )   )   
    

class AreaFrom(object):
    """ Contains constants for area calculation options. """
    # The constant values are arbitrary and may change
    PLOT_1 = 1
    PLOT_2 = 2
    PLOT_3 = 3
    
class BaselineFilterMethod(object):
    """ Baseline Filter Method """
    NONE = 0
    LOWESS = 1
    BSPLINE = 2
    
    # Items for the Voigt fitting tool radio buttons
    choices = ordered_dict(   (   (NONE , "None"), 
                                        (LOWESS , "Lowess"),
                                        (BSPLINE , "B-Spline"),
                                )   )   

class BaselineFilterRange(object):
    """ Baseline Filter Range """
    WHOLE_SPECTRUM = 0
    BETWEEN_CURSORS = 1
    
    # Items for the Voigt fitting tool radio buttons
    choices = ordered_dict(   (   (WHOLE_SPECTRUM , "Whole Spectrum"), 
                                        (BETWEEN_CURSORS , "Between Cursors"),
                                )   )   

class DcOffset(object):
    """ DC offset constants """
    MIN = -1e5
    MAX =  1e5

class EddyCurrentCorrection(object):
    """ Eddy current correction constants """
    # These constants are arbitrary and may change.
    # However bool(NONE) is guaranteed to be False
    NONE = 0
    SIMPLE = 1
    
    MIN_NOISE_FILTER =   0.1
    MAX_NOISE_FILTER = 100.0
   
    # Items for the spectral processing options dropdown
    choices = ordered_dict(   (   (NONE , "Off"), 
                                        (SIMPLE , "Simple"),
                                )   )   


class FittingLineshapeModel(object):
    """ Lineshape model """
    VOIGT = 0
    LORENTZ = 1
    GAUSS = 2
    
    # Items for the Voigt fitting tool radio buttons
    choices = ordered_dict(   (   (VOIGT , "Voigt"), 
                                        (LORENTZ , "Lorentz"),
                                        (GAUSS , "Gauss"),
                                )   )   


class FittingBaselineAlgorithm(object):
    """ Baseline Algorithm """
    NONE = 0
    VARIABLE_KNOT_SPLINE = 1
    FIXED_KNOT_SPLINE = 2
    WAVELET_FILTER = 3
    
    # Items for the Voigt fitting tool radio buttons
    choices = ordered_dict(   (   (NONE , "None"), 
                                        (VARIABLE_KNOT_SPLINE , "Variable Knot Spline"),
                                        (FIXED_KNOT_SPLINE , "Fixed Knot Spline"),
                                        (WAVELET_FILTER , "Wavelet Filter"),
                                )   )   


class FittingMacromoleculePeak(object):
    """ Macromolecular Peak Method """
    GROUPED_PEAKS = 0
    INDIVIDUAL_PEAKS = 1
    
    # Items for the Voigt fitting tool radio buttons
    choices = ordered_dict(   (   (GROUPED_PEAKS , "Grouped Peaks"), 
                                        (INDIVIDUAL_PEAKS , "Individual Peaks"),
                                )   )   


class FittingMacromoleculeLinewidths(object):
    """ Macromolecular Linewidths Method """
    LUMPED = 0
    INDEPENDENT = 1
    
    # Items for the Voigt fitting tool radio buttons
    choices = ordered_dict(   (   (LUMPED , "Lumped"), 
                                        (INDEPENDENT , "Independent"),
                                )   )   


class FittingOptimizeAlgorithm(object):
    """ Optimization Algorithm """
    NONE = 0
    CONSTRAINED_LEVENBERG_MARQUARDT = 1
    
    # Items for the Voigt fitting tool radio buttons
    choices = ordered_dict(   (   (NONE , "None"), 
                                        (CONSTRAINED_LEVENBERG_MARQUARDT , "ccfit"),
                                )   )   


class FittingOptimizeWeights(object):
    """ Macromolecular Linewidths Method """
    EVEN_WEIGHTING = 0
    LOCAL_WEIGHTING = 1
    
    # Items for the Voigt fitting tool radio buttons
    choices = ordered_dict(   (   (EVEN_WEIGHTING , "Even Weighting"), 
                                        (LOCAL_WEIGHTING , "Local Weighting"),
                                )   )   


class FittingAmplitudeMultiplier(object):
    """ Metabolites amplitude multiplier constants """
    MIN =    0.001
    MAX = 1000.0

class FittingBaselineBsplineOrder(object):
    """ Baseline B-Spline order constants """
    MIN = 1.0
    MAX = 5.0

class FittingBaselineLowessWindowSize(object):
    """ Baseline metabolites region Lowess window size (Hz) constants """
    MIN =    0.00001
    MAX = 5000.0

class FittingBaselineUnderestimation(object):
    """ Baseline first pass underestimation constants """
    MIN =  -50.0
    MAX =  100.0

class FittingLineWidth(object):
    """ Metabolites line width constants """
    MIN =    0.001
    MAX = 1000.0
    
class FittingMacroMoleculeLines(object):
    """ Metabolites Lorentz-Gauss macro molecule model lines constants """
    MIN =  1
    MAX = 50
    
class FittingMacroMoleculeAdjustment(object):
    """ Metabolites Lorentz-Gauss macro molecule adjustment (spinner) constants """
    MIN =   0
    MAX = 100
    
class FittingOptimizationAmplitude(object):
    """ Fitting Lorentz-Gauss optimization metabolite amplitude constants """
    MIN =     1
    MAX = 10000

class FittingOptimizationAreaWeight(object):
    """ Fitting Lorentz-Gauss optimization area weight constants """
    MIN =       0.0001
    MAX = 1000000.0

class FittingOptimizationAlgorithmIterations(object):
    """ Fitting Lorentz-Gauss optimization algorithm max iterations constants """
    MIN =     1
    MAX = 10000

class FittingOptimizationConfidenceAlpha(object):
    """ Fitting Lorentz-Gauss optimization algorithm confidence alpha constants """
    MIN = 0.05
    MAX = 0.9999

class FittingOptimizationFrequency(object):
    """ Fitting Lorentz-Gauss optimization metabolite frequency constants """
    MIN =     1
    MAX = 10000

class FittingOptimizationGlobalIterations(object):
    """ Fitting Lorentz-Gauss optimization global iterations constants """
    MIN =    1
    MAX = 1000

class FittingOptimizationLocalMultiplier(object):
    """ Fitting Lorentz-Gauss optimization "LW Local Mult" (???) constants """
    MIN =   1.0
    MAX = 100.0

class FittingOptimizationPhase1(object):
    """ Fitting Lorentz-Gauss optimization metabolite phase 1 constants """
    MIN =    1
    MAX = 5000

class FittingOptimizationStopTolerance(object):
    """ Fitting Lorentz-Gauss optimization algorithm stop tolerance constants """
    MIN =      0.000000001
    MAX = 100000.0

class FittingOptimizationTaTb(object):
    """ Fitting Lorentz-Gauss optimization Ta=Tb constants """
    MIN =   0.001
    MAX = 200.0

class FittingPeakPpm(object):
    """ Metabolites peak PPM constants """
    MIN = -5000.0
    MAX =  5000.0
    
class FittingPeakSearchRange(object):
    """ Metabolites peak search range constants """
    MIN =  0.001
    MAX = 10.0    
    
class FrequencyShift(object):
    """ Frequency shift constants """
    MIN = -1e4
    MAX =  1e4
    
    
class Phase_1(object):
    """ First order phase constants """
    MIN = -1e4
    MAX =  1e4
    
    MIN_PIVOT = -1000
    MAX_PIVOT =  1000
    
class Plot3Function(object):
    """ Contains constants for plot 3 function options. """
    # The constant values are arbitrary and may change.
    # However, bool(NONE) is guaranteed to be False
    NONE = 0
    RESIDUAL_1_MINUS_2 = 1
    RESIDUAL_2_MINUS_1 = 2
    TIME_SERIES_SUM_1 = 3
    TIME_SERIES_SUM_2 = 4

class SpatialFilter(object):
    """ Spatial filter constants """
    # These constants are arbitrary and may change.
    # However bool(NONE) is guaranteed to be False
    NONE = 0
    HAMMING = 1
    EXPONENTIAL = 2
    GAUSSIAN = 3

    # Items for the spatial processing options dropdown
    choices = ordered_dict(   (   (NONE , "None"), 
                                        (HAMMING , "Hamming"),
                                        (EXPONENTIAL , "Exponential"),
                                        (GAUSSIAN , "Gaussian"),
                                )   )   

class SpatialTranspose(object):
    """ Spatial transposition constants """
    # These constants are arbitrary and may change.
    # However bool(NONE) is guaranteed to be False
    NONE = 0
    TRANSPOSE_XY = 1    # x <-> y
    TRANSPOSE_XZ = 2    # x <-> z
    TRANSPOSE_YZ = 3    # y <-> z
    TRANSPOSE_XYZ = 4   # x->y->z->x

    # Items for the spatial processing options
    choices = ordered_dict(   (   (NONE , "None"), 
                                        (TRANSPOSE_XY , "Transpose_XY"),
                                        (TRANSPOSE_XZ , "Transpose_XZ"),
                                        (TRANSPOSE_YZ , "Transpose_YZ"),
                                        (TRANSPOSE_XYZ , "Transpose_XYZ"),
                                )   )

class WaterExtrapolation(object):
    """ Water extrapolation constants """
    # These constants are arbitrary and may change.
    # However bool(NONE) is guaranteed to be False
    NONE = 0
    LINEAR = 1
    AR_MODEL = 2
    
    MIN_POINTS = 1
    MAX_POINTS = 1000

    # Items for the spectral processing options dropdown
    choices = ordered_dict(   (   (NONE , "None"), 
                                        (LINEAR , "Linear"),
                                        (AR_MODEL , "AR Model"),
                                )   )   
    

class WaterFilter(object):
    """ Water filter constants """
    
    # These constants are arbitrary and may change.
    # However bool(NONE) is guaranteed to be False
    NONE = 0
    HAMMING = 1
    FIR = 2
    HLSVD = 3

    # Items for the spectral processing options dropdown
    choices = ordered_dict(   (   (NONE , "None"), 
                                        (HAMMING , "Hamming"),
                                        (FIR , "FIR"),
                                        (HLSVD , "HLSVD"),
                                )   )   

    # FIR (Finite Impulse Response) constants
    MIN_FIR_RIPPLE = 0
    MAX_FIR_RIPPLE = 500
    
    MIN_FIR_HALF_WIDTH =   0
    MAX_FIR_HALF_WIDTH = 500
    
    # HLSVD constants
    MIN_HLSVD_DATA_POINTS =    1
    MAX_HLSVD_DATA_POINTS = 2048
    
    MIN_HLSVD_SINGULAR_VALUES =  1
    MAX_HLSVD_SINGULAR_VALUES = 50    

    MIN_HLSVD_MATRIX_POINTS =    1
    MAX_HLSVD_MATRIX_POINTS = 1024

    MIN_HLSVD_ITERATIONS =    1
    MAX_HLSVD_ITERATIONS = 2000
    

class XYZImageType(object):
    """ XYZ Image Type constants """
    # Integral Image along 0:X-Y/1:Y-Z/2:X-Z direction
    MAGNITUDE = 0
    REAL      = 1
    IMAGINARY = 2
    MAGNITUDE_PLUS_MASK = 3
    B0_MAP    = 4

    # Items for the XYZ browser options
    choices = ordered_dict(   (   (MAGNITUDE , "Magnitude"), 
                                        (REAL , "Real"),
                                        (IMAGINARY , "Imaginary"),
                                        (MAGNITUDE_PLUS_MASK , "Magnitude plus Mask"),
                                        (B0_MAP , "B0 Map"),
                                )   )   


class XYZOrientation(object):
    """ XYZ Orientation constants """
    # Integral Image along 0:X-Y/1:Y-Z/2:X-Z direction
    XY_ORIENTATION = 0
    YZ_ORIENTATION = 1
    XZ_ORIENTATION = 2

    # Items for the XYZ browser options
    choices = ordered_dict(   (   (XY_ORIENTATION , "XY_Orientation"), 
                                        (YZ_ORIENTATION , "YZ_Orientation"),
                                        (XZ_ORIENTATION , "XZ_Orientation"),
                                )   )   


class ZeroFillMultiplier(object):
    """ Zero fill multiplier constants. The zero fill is not only limited
    to a specific range, but it must also be an integral power of 2.
    """
    _EXPONENT = 5
    MIN = 1
    MAX = 2 ** _EXPONENT

    # Items for the spatial processing options
    choices = [ (2 ** i, str(2 ** i)) for i in range(0, _EXPONENT + 1) ]
    choices = ordered_dict(choices)   




"""
Helper routines for reading & writing MRS data in various formats.

Except for the map of internal data types to numpy type strings (which
doesn't require an import of numpy), this code is deliberately ignorant of
numpy. It returns native Python types that are easy to convert into
numpy types.
"""

# Python modules
from __future__ import division
import re
import struct
import os.path
import xdrlib
import exceptions

class FileNotFoundError(exceptions.Exception):
    """ Specific to VASF. Raised when this module can't find a matching data
    file for a params file or vice versa.
    """
    pass


class UnreadableDataError(exceptions.Exception):
    """ Raised when this module can't make sense of data (raw or XDR) due
    to an unexpected format, buffer underrun (less data than expected),
    buffer overrun (more data than expected), etc.
    """
    pass


class IncompleteMetadataError(exceptions.Exception):
    """ Raised when the metadata associated with a dataset doesn't contain
    required information like the data's type or format.
    """
    pass


# Regexes for sniffing VASF file types. The former looks for a word or
# two inside [SQUARE BRACKETS] and the latter looks for <vasf> or <?xml
_vasf_parameters_regex = re.compile(r"[[](\w+\s*\w*)[]]")
_vaxml_regex = re.compile(r"<vasf>|<\?xml")


class MrsFileTypes(object):
    # These constants are arbitrary and may change.
    # However bool(NONE) is guaranteed to be False
    NONE = 0
    VASF_DATA = 1
    VASF_PARAMETERS = 2
    VAXML = 3
    DICOM_SIEMENS = 4


class DataTypes(object):
    """ Internal representation of data type. INTEGER is for 16 bit values,
    LONG for 32 bit values, FLOAT for 32 bit floats, DOUBLE for 64 bit,
    COMPLEX for two 32 bit floats, COMPLEX_DOUBLE for two 64 bit floats.
    """

    # These constants are arbitrary and may change except that bool(NONE) is
    # always guaranteed to be False.
    NONE = 0
    BYTE = 1
    INTEGER = 2
    LONG = 3
    FLOAT = 4
    DOUBLE = 5
    COMPLEX = 6
    COMPLEX_DOUBLE = 7

    ALL = (BYTE, INTEGER, LONG, FLOAT, DOUBLE, COMPLEX, COMPLEX_DOUBLE)

    # Sizes of types, in bytes
    _TYPE_SIZES = {
                    NONE            :    0,
                    BYTE            :    1,
                    INTEGER         :    2,
                    LONG            :    4,
                    FLOAT           :    4,
                    DOUBLE          :    8,
                    COMPLEX         :    8,
                    COMPLEX_DOUBLE  :   16
                }


    # Maps external type strings to internal values. External strings
    # are the many variations one can find in VASF format. They also
    # include the numpy type strings. Older code tended to strip the spaces
    # out of values read from the INI file, so all the type names have to
    # appear in spaceless (e.g. "doublefloat") as well as "spaced" form.
    _EXTERNAL_TO_INTERNAL = {
        # These are VASF strings
        'float'                     :    FLOAT,
        'double'                    :    DOUBLE,
        'doublefloat'               :    DOUBLE,
        'double float'              :    DOUBLE,
        'shortinteger'              :    INTEGER,
        'short integer'             :    INTEGER,
        'integer'                   :    INTEGER,
        'unsignedinteger'           :    INTEGER,
        'unsigned integer'          :    INTEGER,
        'integer16bit'              :    INTEGER,
        'integer 16bit'             :    INTEGER,
        'integer 16 bit'            :    INTEGER,
        'integer'                   :    INTEGER,
        'long'                      :    LONG,
        'unsignedlong'              :    LONG,
        'unsigned long'             :    LONG,
        'complexinteger8bit'        :    COMPLEX,
        'complex integer8bit'       :    COMPLEX,
        'complex integer 8bit'      :    COMPLEX,
        'complex integer 8 bit'     :    COMPLEX,
        'complexinteger16bit'       :    COMPLEX,
        'complex integer16bit'      :    COMPLEX,
        'complex integer 16bit'     :    COMPLEX,
        'complex integer 16 bit'    :    COMPLEX,
        'complexfloat'              :    COMPLEX,
        'complex float'             :    COMPLEX,
        'complex'                   :    COMPLEX,
        'complexdouble'             :    COMPLEX_DOUBLE,
        'complex double'            :    COMPLEX_DOUBLE,
        'byte'                      :    BYTE,
        # These are numpy types
        "character"                 :    BYTE,
        "int16"                     :    INTEGER,
        "int32"                     :    LONG,
        "float32"                   :    FLOAT,
        "float64"                   :    DOUBLE,
        "complex64"                 :    COMPLEX,
        "complex128"                :    COMPLEX_DOUBLE,
    }

    _INTERNAL_TO_NUMPY = {
        # Maps internal types to numpy type strings
        # Valid numpy type names are in numpy.sctypeDict.keys()
        BYTE              :   "byte",
        INTEGER           :   "int16",
        LONG              :   "int32",
        FLOAT             :   "float32",
        DOUBLE            :   "float64",
        COMPLEX           :   "complex64",
        COMPLEX_DOUBLE    :   "complex128",
    }


    @staticmethod
    def is_complex(the_type):
        return the_type in (DataTypes.COMPLEX, DataTypes.COMPLEX_DOUBLE)


    @staticmethod
    def any_type_to_internal(the_type):
        if the_type in DataTypes.ALL:
            pass
            # This is already an internal type
        else:
            if the_type in DataTypes._EXTERNAL_TO_INTERNAL:
                the_type = DataTypes._EXTERNAL_TO_INTERNAL[the_type]
            else:
                raise ValueError, 'Unknown type "%s"' % the_type

        return the_type


    @staticmethod
    def any_type_to_numpy(the_type):
        the_type = DataTypes.any_type_to_internal(the_type)
        return DataTypes._INTERNAL_TO_NUMPY[the_type]



def sniff_file_type(f):
    """ Guesses if a file is a DICOM file, VASF params file, a VASF data file,
    or a VAXML file and returns one of the MrsFileTypes.XXX constants.

    The param f can be a filename (string) or a file object.

    Since VASF data files are just binary glop, they're indistinguishable
    from most other file types and therefore MrsFileTypes.VASF_DATA is
    the default "wild guess" return type.
    """
    file_type = MrsFileTypes.NONE

    close_on_exit = False

    # If f isn't already a file, turn it into one.
    if not hasattr(f, "read"):
        f = open(f, "rb")
        close_on_exit = True

    # Read the first 1k
    s = f.read(1024)

    # Look for telltale strings
    # Per the DICOM specs, a DICOM file starts with 128 reserved bytes
    # followed by "DICM".
    # ref: DICOM spec, Part 10: Media Storage and File Format for Media
    # Interchange, 7.1 DICOM FILE META INFORMATION
    if s[128:132] == "DICM":
        file_type = MrsFileTypes.DICOM_SIEMENS
    elif _vasf_parameters_regex.search(s):
        file_type = MrsFileTypes.VASF_PARAMETERS
    elif _vaxml_regex.search(s):
        file_type = MrsFileTypes.VAXML
    else:
        file_type = MrsFileTypes.VASF_DATA

    if close_on_exit:
        f.close()

    return file_type




def decode_xdr(data, data_type, element_count):
    """ Given a string of data in XDR format and a data type, returns
    an iterable (tuple or list) of Python objects representing the decoded
    data. data_type must be one of the DataTypes.XXX constants defined in
    this module.
    element_count is the number of elements expected in the data.
    """
    p = xdrlib.Unpacker(data)

    is_complex = (data_type in (DataTypes.COMPLEX, DataTypes.COMPLEX_DOUBLE))

    if data_type in (DataTypes.COMPLEX, DataTypes.FLOAT):
        unpack_function = p.unpack_float
    elif data_type in (DataTypes.COMPLEX_DOUBLE, DataTypes.DOUBLE):
        unpack_function = p.unpack_double
    elif data_type == DataTypes.LONG:
        unpack_function = p.unpack_int
    elif data_type == DataTypes.INTEGER:
        unpack_function = p.unpack_int
    elif data_type == DataTypes.BYTE:
        unpack_function = p.unpack_byte
    else:
        raise ValueError, "Unknown data type '%s'" % data_type

    if is_complex:
        # XDR doesn't explicitly support complex numbers, so they're written
        # as pairs of floats (or doubles).
        element_count *= 2

    try:
        data = p.unpack_farray(element_count, unpack_function)
    except (xdrlib.Error, xdrlib.ConversionError), instance:
        raise UnreadableDataError, instance.msg

    # Calling p.done() here will raise an xdrlib.Error if unextracted
    # data remains (i.e. the code above is buggy or element_count is wrong)
    try:
        p.done()
    except xdrlib.Error:
        raise UnreadableDataError, "More data in file than expected (XDR overrun)"

    if is_complex:
        data = collapse_complexes(data)

    return data



def collapse_complexes(data):
    """Given a list or other iterable that's a series of (real, imaginary)
    pairs, returns a list of complex numbers. For instance, this list --
       [a, b, c, d, e, f]
    this function returns --
       [complex(a, b), complex(c, d), complex(e, f)]

    The returned list is a new list; the original is unchanged.
    """
    return [complex(data[i], data[i+1]) for i in range(0, len(data), 2)]


def expand_complexes(data):
    """Expands a list or other iterable of complex numbers into a list of
    (real, imaginary) pairs. For instance, given this list of complex
    numbers --
       [za, zb, zc]
    this function returns --
       [za.real, za.imag, zb.real, zb.imag, zc.real, zc.imag]

    The returned list is a new list; the original is unchanged.
    """
    # First I double the length of the list by adding empty elements to
    # the front.
    data = ([None] * len(data)) + list(data)

    # Now I overwrite the items in the list with item N being split
    # into real and imag and list[N/2] = real and list[N/2 + 1] = imag.
    j = 0
    for i in range(len(data) // 2, len(data)):
        data[j] = data[i].real
        j += 1
        data[j] = data[i].imag
        j += 1

    return data


def convert_vasf_to_xml(source_filename, target_filename, pack_data=True):
    """ Converts a VASF file pair to VAXML format.
    source_filename can be a VASF parameters or data filename.
    """
    # imports are done on-the-fly here because util_vasf_file and 
    # util_vaxml_file import this file and we need to avoid circular imports.
    import util_vasf_file
    import util_vaxml_file
    
    source_parameters_filename, source_data_filename = \
                            util_vasf_file.get_filename_pair(source_filename)

    parameters, data = util_vasf_file.read(source_parameters_filename,
                                           source_data_filename)

    # Extract the data type
    util_vaxml_file.write(parameters, parameters["data_type"], data, 
                          target_filename, pack_data)


def _test_collapse_expand_complexes():
    import random
    import numpy

    random.seed()
    LIST_SIZE = random.randint(0, 1000)

    collapsed = [ ]
    raw = [ ]
    # Generate a bunch of random floats
    for i in range(LIST_SIZE):
        real = random.randint(-1000, 1000) + random.random()
        imaginary = random.randint(-1000, 1000) + random.random()

        raw.append(real)
        raw.append(imaginary)
        collapsed.append(complex(real, imaginary))

    assert(collapse_complexes(raw) == collapsed)
    assert(expand_complexes(collapsed) == raw)

    # Ensure the functions work with numpy arrays
    raw = numpy.array(raw)
    collapsed = numpy.array(collapsed)
    assert((numpy.array(collapse_complexes(raw)) == collapsed).all())
    assert((numpy.array(expand_complexes(collapsed)) == raw).all())


if __name__ == '__main__':
    _test_collapse_expand_complexes()




#!/usr/bin/env python

import os
from setuptools import setup

def get_readme():
    md_path = os.path.join(os.path.dirname(__file__), "README.md")
    txt_path = os.path.join(os.path.dirname(__file__), "README.txt")
    if os.path.exists(txt_path):
        d = open(txt_path).read()
    elif os.path.exists(md_path):
        d = open(md_path).read()
    else:
        d = ""
    return d

setup(name='hcpre',
      version='0.5.0',
      author='Ben Acland',
      author_email='benacland@gmail.com',
      description='Generalized launcher for human connectome project BOLD preprocessing',
      license='BSD',
      keywords='connectome preprocessing fmri bold',
      url='https://github.com/beOn/hcpre',
      setup_requires=['nibabel','traits','networkx','pydicom'],
      install_requires=['pydicom','nibabel','networkx','traits','nipype','configobj'],
      packages=['hcpre'],
      scripts=['hcpre/hcpre.py'],
      package_data={'hcpre':['duke_siemens/*']},
      long_description=get_readme(),
      classifiers=[
        'Development Status :: 3 - Alpha',
        'Topic :: Scientific/Engineering :: Information Analysis',
        'License :: OSI Approved :: BSD License',
        'Intended Audience :: Science/Research',
      ],
)
